import os
import copy
import numpy as np

import cloudvolume as cv
from cloudvolume.lib import Bbox

from torch.nn.functional import interpolate

from corgie import layers
from corgie import exceptions
from corgie.log import logger as corgie_logger

from corgie.mipless_cloudvolume import MiplessCloudVolume
from corgie.data_backends.base import (
    DataBackendBase,
    BaseLayerBackend,
    register_backend,
    str_to_backend,
)


@register_backend("cv")
class CVDataBackend(DataBackendBase):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)


cv_backend = str_to_backend("cv")


class CVLayerBase(BaseLayerBackend):
    def __init__(
        self,
        path,
        backend,
        reference=None,
        force_chunk_xy=None,
        force_chunk_z=None,
        overwrite=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.path = path
        self.backend = backend

        try:
            info = MiplessCloudVolume(path).get_info()
        except cv.exceptions.InfoUnavailableError as e:
            if reference is None:
                raise e
            else:
                info = copy.deepcopy(reference.get_info())
                overwrite = True

        if overwrite:
            info["num_channels"] = self.get_num_channels()
            if self.dtype is None:
                dtype = self.get_default_data_type()
            else:
                dtype = self.dtype
            info["data_type"] = dtype
            if not self.supports_voxel_offset():
                for scale in info["scales"]:
                    scale["voxel_offset"] = [0, 0, 0]
            if not self.supports_chunking():
                for scale in info["scales"]:
                    scale["chunk_sizes"] = [[1, 1, 1]]
            if force_chunk_z is not None:
                for scale in info["scales"]:
                    scale["chunk_sizes"][0][-1] = force_chunk_z
            if force_chunk_xy is not None:
                for scale in info["scales"]:
                    scale["chunk_sizes"][0][0] = force_chunk_xy
                    scale["chunk_sizes"][0][1] = force_chunk_xy

        self.cv = MiplessCloudVolume(path, info=info, overwrite=overwrite, **kwargs)

        self.dtype = info["data_type"]

    def __str__(self):
        return "CV {}".format(self.path)

    def __repr__(self):
        return self.__str__()

    def get_sublayer(self, name, layer_type=None, path=None, **kwargs):
        if path is None:
            path = os.path.join(self.cv.path, layer_type, name)

        if layer_type is None:
            layer_type = self.get_layer_type()

        return self.backend.create_layer(
            path=path, reference=self, layer_type=layer_type, **kwargs
        )

    def read_backend(self, bcube, mip, transpose=True, timestamp=None, **kwargs):
        x_range = bcube.x_range(mip)
        y_range = bcube.y_range(mip)
        z_range = bcube.z_range()

        this_cv = self.cv[mip]
        x_off, y_off, z_off = this_cv.voxel_offset
        """if x_range[0] < x_off:
            corgie_logger.debug(f"READ from {str(self)}: \n"
                    f"   reducing xs from {x_range[0]} to {x_off} MIP: {mip}")
            x_range[0] = x_off
        if y_range[0] < y_off:
            corgie_logger.debug(f"READ from {str(self)}: \n"
                    f"   reducing ys from {y_range[0]} to {y_off} MIP: {mip}")
            y_range[0] = y_off
        if z_range[0] < z_off:
            corgie_logger.debug(f"READ from {str(self)}: \n"
                    f"   reducing zs from {z_range[0]} to {z_off} MIP: {mip}")
            z_range[0] = z_off"""

        corgie_logger.debug(
            "READ from {}: \n   x: {}, y: {}, z: {}, MIP: {}".format(
                str(self), x_range, y_range, z_range, mip
            )
        )
        if timestamp is not None:
            data = self.cv[mip].download(
                bbox=(
                    slice(x_range[0], x_range[1], None),
                    slice(y_range[0], y_range[1], None),
                    slice(z_range[0], z_range[1], None),
                ),
                mip=mip,
                preserve_zeros=True,
                parallel=self.cv[mip].parallel,
                agglomerate=True,
                timestamp=timestamp,
            )
        else:
            data = self.cv[mip][
                x_range[0] : x_range[1],
                y_range[0] : y_range[1],
                z_range[0] : z_range[1],
            ]
        if transpose:
            data = np.transpose(data, (2, 3, 0, 1))
        return data

    def write_backend(self, data, bcube, mip):
        x_range = bcube.x_range(mip)
        y_range = bcube.y_range(mip)
        z_range = bcube.z_range()

        data = np.transpose(data, (2, 3, 0, 1))
        corgie_logger.debug(
            "Write to {}: \n x: {}, y: {}, z: {}, MIP: {}".format(
                str(self), x_range, y_range, z_range, mip
            )
        )
        self.cv[mip].autocrop = True
        self.cv[mip][
            x_range[0] : x_range[1], y_range[0] : y_range[1], z_range[0] : z_range[1]
        ] = data
        self.cv[mip].autocrop = False

    def get_info(self):
        return self.cv.get_info()

    def resolution(self, mip):
        return self.cv[mip].resolution

    def get_chunk_aligned_bcube(self, bcube, mip, chunk_xy, chunk_z):
        cv_chunk = self.cv[mip].chunk_size

        # Expand bbox to be difizible by chunk_size

        bbox = Bbox(
            (bcube.x_range(mip)[0], bcube.y_range(mip)[0], bcube.z_range()[0]),
            (bcube.x_range(mip)[1], bcube.y_range(mip)[1], bcube.z_range()[1]),
        )

        aligned_bbox = bbox.expand_to_chunk_size(
            self.cv[mip].chunk_size, self.cv[mip].voxel_offset
        )

        aligned_bcube = copy.deepcopy(bcube)
        aligned_bcube.reset_coords(
            aligned_bbox.minpt[0],
            aligned_bbox.maxpt[0],
            aligned_bbox.minpt[1],
            aligned_bbox.maxpt[1],
            aligned_bbox.minpt[2],
            aligned_bbox.maxpt[2],
            mip=mip,
        )

        if chunk_xy % cv_chunk[0] != 0:
            raise exceptions.ChunkingError(
                self,
                f"Processing chunk_xy {chunk_xy} is not"
                f"divisible by MIP{mip} CV chunk {cv_chunk[0]}",
            )
        if chunk_z % cv_chunk[2] != 0:
            raise exceptions.ChunkingError(
                self,
                f"Processing chunk_z {chunk_z} is not"
                f"divisible by MIP{mip} CV chunk {cv_chunk[2]}",
            )

        if chunk_xy > aligned_bcube.x_size(mip):
            x_adj = chunk_xy - aligned_bcube.x_size(mip)
        else:
            x_rem = aligned_bcube.x_size(mip) % chunk_xy
            if x_rem == 0:
                x_adj = 0
            else:
                x_adj = chunk_xy - x_rem
        if chunk_xy > aligned_bcube.y_size(mip):
            y_adj = chunk_xy - aligned_bcube.y_size(mip)
        else:
            y_rem = aligned_bcube.y_size(mip) % chunk_xy
            if y_rem == 0:
                y_adj = 0
            else:
                y_adj = chunk_xy - y_rem

        if chunk_z > aligned_bcube.z_size():
            z_adj = chunk_z - aligned_bcube.z_size()
        else:
            z_rem = aligned_bcube.z_size() % chunk_z
            if z_rem == 0:
                z_adj = 0
            else:
                z_adj = chunk_z - z_rem
        if x_adj != 0:
            xe = aligned_bcube.x_range(mip)[1]
            aligned_bcube.reset_coords(xe=xe + x_adj, mip=mip)
        if y_adj != 0:
            ye = aligned_bcube.y_range(mip)[1]
            aligned_bcube.reset_coords(ye=ye + y_adj, mip=mip)
        if z_adj != 0:
            ze = aligned_bcube.z_range()[1]
            aligned_bcube.reset_coords(ze=ze + z_adj)
        return aligned_bcube

    def break_bcube_into_chunks(
        self, bcube, chunk_xy, chunk_z, mip, readonly=False, **kwargs
    ):
        if not readonly:
            bcube = self.get_chunk_aligned_bcube(bcube, mip, chunk_xy, chunk_z)

        chunks = super().break_bcube_into_chunks(
            bcube, chunk_xy, chunk_z, mip, **kwargs
        )
        return chunks

    def flush(self, mip):
        corgie_logger.debug(f"FLUSH {str(self)} at {mip}")
        self.cv[mip].cache.flush()


@cv_backend.register_layer_type_backend("img")
class CVImgLayer(CVLayerBase, layers.ImgLayer):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)


@cv_backend.register_layer_type_backend("segmentation")
class CVSegmentationLayer(CVLayerBase, layers.SegmentationLayer):
    def __init__(self, *kargs, **kwargs):
        if "graphene:" in kwargs["path"]:
            kwargs["readonly"] = True
        super().__init__(*kargs, **kwargs)


@cv_backend.register_layer_type_backend("field")
class CVFieldLayer(CVLayerBase, layers.FieldLayer):
    supported_backend_dtypes = ["float32", "int16"]

    def __init__(self, backend_dtype="float32", **kwargs):
        super().__init__(**kwargs)
        if backend_dtype not in self.supported_backend_dtypes:
            raise exceptions.ArgumentError(
                "Field layer 'backend_dtype'",
                "\n{} is not a supported field backend data type. \n"
                "Supported backend data types: {}".format(
                    backend_dtype, self.supported_backend_dtypes
                ),
            )
        self.backend_dtype = backend_dtype

    def read_backend(self, bcube, mip, transpose=True, timestamp=None, **kwargs):
        data = super().read_backend(bcube, mip, transpose, timestamp, **kwargs)
        if self.backend_dtype == "int16":
            data = np.float32(data) / 4
        return data


@cv_backend.register_layer_type_backend("mask")
class CVMaskLayer(CVLayerBase, layers.MaskLayer):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)


@cv_backend.register_layer_type_backend("section_value")
class CVSectionValueLayer(CVLayerBase, layers.SectionValueLayer):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)


@cv_backend.register_layer_type_backend("fixed_field")
class CVFixedFieldLayer(CVFieldLayer, layers.FixedFieldLayer):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)


@cv_backend.register_layer_type_backend("float_tensor")
class CVFloatTensorLayer(CVLayerBase, layers.FloatTensorLayer):
    def __init__(self, *kargs, **kwargs):
        super().__init__(*kargs, **kwargs)



import click
import procspec
import json

from corgie import scheduling, argparsers, helpers

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

from corgie.cli.common import ChunkedJob
from corgie.cli.downsample import DownsampleJob


class ApplyProcessorJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        dst_layer,
        chunk_xy,
        chunk_z,
        processor_spec,
        processor_mip,
        processor_mip_out,
        pad,
        crop,
        bcube,
        blend_xy=0,
    ):

        self.src_stack = src_stack
        self.dst_layer = dst_layer
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.blend_xy = blend_xy
        self.pad = pad
        self.crop = crop
        self.bcube = bcube
        self.processor_spec = processor_spec
        self.processor_mip = processor_mip
        self.processor_mip_out = processor_mip_out


        if isinstance(self.processor_spec, str):
            self.processor_spec = [self.processor_spec]
        if isinstance(self.processor_mip, int):
            self.processor_mip = [self.processor_mip]

        if self.processor_mip_out is None or len(self.processor_mip_out) == 0:
            self.processor_mip_out = self.processor_mip

        if len(self.processor_mip) != len(self.processor_spec):
            raise exceptions.CorgieException(
                "The number of processors doesn't "
                "match the number of specified processor MIPs"
            )

        super().__init__()

    def task_generator(self):
        for i in range(len(self.processor_spec)):
            this_proc = self.processor_spec[i]
            this_proc_mip = self.processor_mip[i]
            this_proc_mip_out = self.processor_mip_out[i]
            is_last_proc = i == len(self.processor_spec) - 1

            this_task = helpers.PartialSpecification(
                ApplyProcessorTask,
                src_stack=self.src_stack,
                processor_spec=this_proc,
                pad=self.pad,
                crop=self.crop,
                mip_in=this_proc_mip,
            )

            chunked_job = ChunkedJob(
                task_class=this_task,
                dst_layer=self.dst_layer,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                blend_xy=self.blend_xy,
                mip=this_proc_mip_out,
                bcube=self.bcube,
            )

            yield from chunked_job.task_generator

            if not is_last_proc:
                yield scheduling.wait_until_done
                next_proc_mip = self.processor_mip[i + 1]
                if this_proc_mip > next_proc_mip:
                    downsample_job = DownsampleJob(
                        src_layer=self.dst_layer,
                        chunk_xy=self.chunk_xy,
                        chunk_z=self.chunk_z,
                        mip_start=this_proc_mip,
                        mip_end=next_proc_mip,
                        bcube=self.bcube,
                    )
                    yield from downsample_job.task_generator
                    yield scheduling.wait_until_done

        if self.processor_mip[0] > self.processor_mip[-1]:
            # good manners
            # prepare the ground for the next you
            downsample_job = DownsampleJob(
                src_layer=self.dst_layer,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip_start=self.processor_mip[-1],
                mip_end=self.processor_mip[0],
                bcube=self.bcube,
            )


class ApplyProcessorTask(scheduling.Task):
    def __init__(self, processor_spec, src_stack, dst_layer, mip, pad, crop, bcube, mip_in):
        super().__init__()
        self.processor_spec = processor_spec
        self.src_stack = src_stack
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.crop = crop
        self.bcube = bcube
        self.mip_in = mip_in

    def execute(self):
        src_bcube = self.bcube.uncrop(self.pad, self.mip_in)

        processor = procspec.parse_proc(spec_str=self.processor_spec)

        src_translation, src_data_dict = self.src_stack.read_data_dict(
            src_bcube, mip=self.mip_in, stack_name="src"
        )

        processor_input = {**src_data_dict}
        result = processor(processor_input, output_key="output")
        cropped_result = helpers.crop(result, self.crop // (2**(self.mip - self.mip_in)))
        self.dst_layer.write(cropped_result, bcube=self.bcube, mip=self.mip)


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
#
@corgie_option(
    "--dst_layer_spec",
    nargs=1,
    type=str,
    required=True,
    help="Specification for the destination layer. Must be a field type.",
)
@corgie_option(
    "--spec_path",
    nargs=1,
    type=str,
    required=False,
    help="JSON spec relating src stacks, src z to dst z",
)
@corgie_option("--reference_key", nargs=1, type=str, default="img")
@corgie_optgroup("Apply Processor Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--force_chunk_xy", nargs=1, type=int, default=None)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option("--blend_xy", nargs=1, type=int, default=0)
@corgie_option(
    "--pad", nargs=1, type=int, default=512,
)
@corgie_option("--crop", nargs=1, type=int, default=None)
@corgie_option("--processor_spec", nargs=1, type=str, multiple=True, required=True)
@corgie_option("--processor_mip", nargs=1, type=int, multiple=True, required=True)
@corgie_option("--processor_mip_out", nargs=1, type=int, multiple=True, required=False)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def apply_processor(
    ctx,
    src_layer_spec,
    dst_layer_spec,
    spec_path,
    processor_spec,
    pad,
    crop,
    chunk_xy,
    start_coord,
    force_chunk_xy,
    processor_mip,
    processor_mip_out,
    end_coord,
    coord_mip,
    blend_xy,
    chunk_z,
    reference_key,
):
    scheduler = ctx.obj["scheduler"]

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    reference_layer = None
    if reference_key in src_stack.layers:
        reference_layer = src_stack.layers[reference_key]

    if force_chunk_xy is None:
        force_chunk_xy = chunk_xy

    dst_layer = create_layer_from_spec(
        dst_layer_spec,
        allowed_types=["img", "mask", "section_value", "field"],
        default_type="img",
        readonly=False,
        caller_name="dst_layer",
        force_chunk_xy=force_chunk_xy,
        reference=reference_layer,
        overwrite=True,
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    if crop is None:
        crop = pad

    if spec_path:
        with open(spec_path, "r") as f:
            spec = json.load(f)

        for z in spec:
            job_bcube = bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            apply_processor_job = ApplyProcessorJob(
                src_stack=src_stack,
                dst_layer=dst_layer,
                chunk_xy=chunk_xy,
                chunk_z=chunk_z,
                blend_xy=blend_xy,
                processor_spec=processor_spec,
                pad=pad,
                crop=crop,
                bcube=job_bcube,
                processor_mip=processor_mip,
                processor_mip_out=processor_mip_out,
            )

            # create scheduler and execute the job
            scheduler.register_job(
                apply_processor_job, job_name="Apply Processor {}".format(job_bcube)
            )
    else:
        apply_processor_job = ApplyProcessorJob(
            src_stack=src_stack,
            dst_layer=dst_layer,
            chunk_xy=chunk_xy,
            chunk_z=chunk_z,
            blend_xy=blend_xy,
            processor_spec=processor_spec,
            pad=pad,
            crop=crop,
            bcube=bcube,
            processor_mip=processor_mip,
            processor_mip_out=processor_mip_out,
        )

        # create scheduler and execute the job
        scheduler.register_job(
            apply_processor_job, job_name="Apply Processor {}".format(bcube)
        )
    scheduler.execute_until_completion()



import click
import os
import math
from corgie.block import get_blocks
from copy import deepcopy

from corgie import exceptions, stack, helpers, scheduling, argparsers

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

from corgie.cli.align_block import AlignBlockJob
from corgie.cli.render import RenderJob
from corgie.cli.copy import CopyLayerJob
from corgie.cli.downsample import DownsampleJob
from corgie.cli.compute_field import ComputeFieldJob
from corgie.cli.seethrough import SeethroughCompareJob
from corgie.cli.vote import VoteJob
from corgie.cli.broadcast import BroadcastJob


@click.command()
# Layers
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@corgie_optgroup("Render Method Specification")
# @corgie_option('--seethrough_masks',    nargs=1, type=bool, default=False)
# @corgie_option('--seethrough_misalign', nargs=1, type=bool, default=False)
@corgie_option('--align_blocks/--no_align_blocks', default=True)
@corgie_option('--compute_stitch_fields/--no_compute_stitch_fields', default=True)
@corgie_option('--compose_fields/--no_compose_fields', default=True)
@corgie_option('--final_render/--no_final_render', default=True)
@corgie_option("--render_pad", nargs=1, type=int, default=512)
@corgie_option("--render_chunk_xy", nargs=1, type=int, default=1024)
@corgie_optgroup("Compute Field Method Specification")
@corgie_option("--processor_spec", nargs=1, type=str, required=True, multiple=True)
@corgie_option("--processor_mip", "-m", nargs=1, type=int, required=True, multiple=True)
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--blend_xy", nargs=1, type=int, default=0)
@corgie_option("--force_chunk_xy", nargs=1, type=int, default=None)
@corgie_option("--pad", nargs=1, type=int, default=256)
@corgie_option("--crop", nargs=1, type=int, default=None)
@corgie_option(
    "--seethrough_spec",
    nargs=1,
    type=str,
    default=None,
    multiple=True,
    help="Seethrough method spec. Use multiple times to specify different methods (e.g. seethrough misalignments, seethrough black, etc.)",
)
@corgie_option(
    "--seethrough_limit",
    nargs=1,
    type=int,
    default=None,
    multiple=True,
    help="For each seethrough method, how many sections are allowed to be seenthrough. 0 or None means no limit.",
)
@corgie_option("--seethrough_spec_mip", nargs=1, type=int, default=None)
@corgie_optgroup("Voting Specification")
@corgie_option("--consensus_threshold", nargs=1, type=float, default=3.)
@corgie_option("--blur_sigma", nargs=1, type=float, default=15.0)
@corgie_option("--kernel_size", nargs=1, type=int, default=32)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@corgie_optgroup("Block Alignment Specification")
@corgie_option(
    "--bad_starter_path",
    nargs=1,
    type=str,
    default=None,
    help="Filepath to list of sections that should not be used as the first section of any block.",
)
@corgie_option("--block_size", nargs=1, type=int, default=10)
@corgie_option(
    "--stitch_size",
    nargs=1,
    type=int,
    default=3,
    help="The number of sections used to compute stitching fields. If >1, then multiple stitching field estimates will be made, then corrected by voting.",
)
@corgie_option(
    "--vote_dist",
    nargs=1,
    type=int,
    default=1,
    help="The number of previous sections for which fields will be estimated, then corrected by voting.",
)
@corgie_option(
    "--decay_dist",
    nargs=1,
    type=int,
    default=100,
    help="The maximum distance in sections over which a previous section may influence a later one.",
)
@corgie_option(
    "--blur_rate",
    nargs=1,
    type=float,
    default=0.2,
    help="The increase in the size of downsample factor based on distance used in broadcasting a stitching field.",
)
@corgie_option(
    "--restart_stage",
    nargs=1,
    type=int,
    default=0,
    help="0: align blocks, 1: align overlaps, 2: vote over overlaps, 3: broadcast stitch field, 4: render with composed field",
)
@corgie_option(
    "--restart_suffix",
    nargs=1,
    type=str,
    default=None,
    help="The suffix of a previous alignment run, which will be used as a starting point. New layers created after the restart will be labeld with --suffix",
)
@click.pass_context
def align(
    ctx,
    src_layer_spec,
    dst_folder,
    render_pad,
    render_chunk_xy,
    processor_spec,
    pad,
    crop,
    processor_mip,
    chunk_xy,
    start_coord,
    end_coord,
    coord_mip,
    bad_starter_path,
    block_size,
    stitch_size,
    vote_dist,
    consensus_threshold,
    blur_sigma,
    kernel_size,
    blend_xy,
    force_chunk_xy,
    suffix,
    seethrough_spec,
    seethrough_limit,
    seethrough_spec_mip,
    decay_dist,
    blur_rate,
    restart_stage,
    restart_suffix,
    align_blocks,
    compute_stitch_fields,
    compose_fields,
    final_render,
):

    scheduler = ctx.obj["scheduler"]

    if suffix is None:
        suffix = "_aligned"
    else:
        suffix = f"_{suffix}"
    if (restart_suffix is None) or (restart_stage == 0):
        restart_suffix = suffix

    if crop is None:
        crop = pad

    corgie_logger.debug("Setting up layers...")

    # TODO: store stitching images in layer other than even & odd
    if vote_dist + stitch_size - 2 >= block_size:
        raise exceptions.CorgieException(
            "block_size too small for stitching + voting requirements (stitch_size + vote_dist)"
        )

    corgie_logger.debug("Setting up layers...")

    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)
    src_stack.folder = dst_folder

    if force_chunk_xy is None:
        force_chunk_xy = chunk_xy

    dst_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=dst_folder,
        name="dst",
        types=["img", "mask"],
        readonly=False,
        suffix=restart_suffix,
        force_chunk_xy=force_chunk_xy,
        overwrite=True,
    )

    even_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=os.path.join(dst_folder, "even"),
        name="even",
        types=["img", "mask"],
        readonly=False,
        suffix=suffix,
        force_chunk_xy=force_chunk_xy,
        overwrite=True,
    )

    odd_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=os.path.join(dst_folder, "odd"),
        name="odd",
        types=["img", "mask"],
        readonly=False,
        suffix=suffix,
        force_chunk_xy=force_chunk_xy,
        overwrite=True,
    )

    corgie_logger.debug("Done!")

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    corgie_logger.debug("Calculating blocks...")
    skip_list = []
    if bad_starter_path is not None:
        with open(bad_starter_path) as f:
            line = f.readline()
            while line:
                skip_list.append(int(line))
                line = f.readline()
    blocks = get_blocks(
        start=bcube.z_range()[0],
        stop=bcube.z_range()[1],
        block_size=block_size,
        block_overlap=1,
        skip_list=skip_list,
        src_stack=src_stack,
        even_stack=even_stack,
        odd_stack=odd_stack,
    )
    stitch_blocks = [b.overlap(stitch_size) for b in blocks[1:]]
    corgie_logger.debug("All Blocks")
    for block, stitch_block in zip(blocks, [None] + stitch_blocks):
        corgie_logger.debug(block)
        corgie_logger.debug(f"Stitch {stitch_block}")
        corgie_logger.debug("\n")

    max_blur_mip = (
        math.ceil(math.log(decay_dist * blur_rate + 1, 2)) + processor_mip[-1]
    )
    corgie_logger.debug(f"Max blur mip for stitching field: {max_blur_mip}")

    # Set all field names, adjusting for restart suffix
    block_field_name = f"field{suffix}"
    stitch_estimated_suffix = f"_stitch_estimated{suffix}"
    stitch_estimated_name = f"field{stitch_estimated_suffix}"
    stitch_corrected_name = f"stitch_corrected{suffix}"
    stitch_corrected_field = None
    composed_name = f"composed{suffix}"
    if restart_stage <= 2:
        stitch_estimated_suffix = f"_stitch_estimated{restart_suffix}"
        stitch_estimated_name = f"field{stitch_estimated_suffix}"
        stitch_corrected_name = f"stitch_corrected{restart_suffix}"
    if restart_stage <= 3:
        composed_name = f"composed{restart_suffix}"

    render_method = helpers.PartialSpecification(
        f=RenderJob,
        pad=render_pad,
        chunk_xy=render_chunk_xy,
        chunk_z=1,
        render_masks=False,
    )

    cf_method = helpers.PartialSpecification(
        f=ComputeFieldJob,
        pad=pad,
        crop=crop,
        processor_mip=processor_mip,
        processor_spec=processor_spec,
        chunk_xy=chunk_xy,
        blend_xy=blend_xy,
        chunk_z=1,
    )
    if seethrough_spec != tuple():
        assert seethrough_spec_mip is not None
        seethrough_method = helpers.PartialSpecification(
            f=SeethroughCompareJob,
            mip=seethrough_spec_mip,
            processor_spec=seethrough_spec,
            chunk_xy=chunk_xy,
            pad=pad,
            crop=pad,
            seethrough_limit=seethrough_limit,
        )
    else:
        seethrough_method = None

    #restart_stage = 4
    #import pdb; pdb.set_trace()
    if restart_stage == 0 and align_blocks:
        corgie_logger.debug("Aligning blocks...")
        for block in blocks:
            block_bcube = block.get_bcube(bcube)
            # Use copies of src & dst so that aligning the stitching blocks
            # is not affected by these block fields.
            # Copying also allows local compute to not modify objects for other tasks
            align_block_job_forv = AlignBlockJob(
                src_stack=deepcopy(block.src_stack),
                dst_stack=deepcopy(block.dst_stack),
                bcube=block_bcube,
                render_method=render_method,
                cf_method=cf_method,
                vote_dist=vote_dist,
                seethrough_method=seethrough_method,
                suffix=suffix,
                copy_start=True,
                use_starters=True,
                backward=False,
                consensus_threshold=consensus_threshold,
                blur_sigma=blur_sigma,
                kernel_size=kernel_size,
            )
            scheduler.register_job(
                align_block_job_forv, job_name=f"Forward Align {block} {block_bcube}",
            )

        scheduler.execute_until_completion()
        corgie_logger.debug("Done!")

    if restart_stage <= 1 and compute_stitch_fields:
        corgie_logger.debug("Aligning stitching blocks...")
        for stitch_block in stitch_blocks:
            block_bcube = stitch_block.get_bcube(bcube)
            # These blocks will have block-aligned images, but not
            # the block_fields that warped them.
            align_block_job_forv = AlignBlockJob(
                src_stack=deepcopy(stitch_block.src_stack),
                dst_stack=deepcopy(stitch_block.dst_stack),
                bcube=block_bcube,
                render_method=render_method,
                cf_method=cf_method,
                vote_dist=vote_dist,
                seethrough_method=seethrough_method,
                suffix=stitch_estimated_suffix,
                copy_start=False,
                use_starters=False,
                backward=False,
                consensus_threshold=consensus_threshold,
                blur_sigma=blur_sigma,
                kernel_size=kernel_size,
            )
            scheduler.register_job(
                align_block_job_forv,
                job_name=f"Stitch Align {stitch_block} {block_bcube}",
            )

        scheduler.execute_until_completion()
        corgie_logger.debug("Done!")

    # Add in the stitch_estimated fields that were just created above
    even_stack.create_sublayer(
        stitch_estimated_name, layer_type="field", overwrite=False,
    )
    odd_stack.create_sublayer(
        stitch_estimated_name, layer_type="field", overwrite=False,
    )
    if restart_stage <= 2 and compose_fields:
        if stitch_size > 1:
            corgie_logger.debug("Voting over stitching blocks")
            stitch_corrected_field = dst_stack.create_sublayer(
                stitch_corrected_name, layer_type="field", overwrite=True
            )
            for stitch_block in stitch_blocks:
                stitch_estimated_field = stitch_block.dst_stack[stitch_estimated_name]
                block_bcube = bcube.reset_coords(
                    zs=stitch_block.start, ze=stitch_block.start + 1, in_place=False,
                )
                z_offsets = [z - block_bcube.z_range()[0] for z in range(stitch_block.start, stitch_block.stop)]
                vote_stitch_job = VoteJob(
                    input_fields=[stitch_estimated_field],
                    output_field=stitch_corrected_field,
                    chunk_xy=chunk_xy,
                    bcube=block_bcube,
                    z_offsets=z_offsets,
                    mip=processor_mip[-1],
                    consensus_threshold=consensus_threshold,
                    blur_sigma=blur_sigma,
                    kernel_size=kernel_size,
                )
                scheduler.register_job(
                    vote_stitch_job,
                    job_name=f"Stitching Vote {stitch_block} {block_bcube}",
                )

            scheduler.execute_until_completion()
            corgie_logger.debug("Done!")

        for stitch_block in stitch_blocks:
            block_bcube = bcube.reset_coords(
                zs=stitch_block.start, ze=stitch_block.start + 1, in_place=False
            )
            field_to_downsample = stitch_block.dst_stack[stitch_estimated_name]
            if stitch_corrected_field is not None:
                field_to_downsample = stitch_corrected_field
            # Hack for fafb
            field_info = field_to_downsample.get_info()
            for scale in field_info['scales']:
                scale['chunk_sizes'][-1][-1] = 1
                scale['encoding'] = 'raw'
            field_to_downsample.cv.store_info(field_info)
            field_to_downsample.cv.fetch_info()
            downsample_field_job = DownsampleJob(
                src_layer=field_to_downsample,
                mip_start=processor_mip[-1],
                mip_end=max_blur_mip,
                bcube=block_bcube,
                chunk_xy=chunk_xy,  # TODO: This probably needs to be modified at highest mips
                chunk_z=1,
                mips_per_task=2,
            )
            scheduler.register_job(
                downsample_field_job,
                job_name=f"Downsample stitching field {block_bcube}",
            )
        scheduler.execute_until_completion()
        corgie_logger.debug("Done!")

    # Add in the block-align fields
    even_stack.create_sublayer(
        block_field_name, layer_type="field", overwrite=False,
    )
    odd_stack.create_sublayer(
        block_field_name, layer_type="field", overwrite=False,
    )
    composed_field = dst_stack.create_sublayer(
        composed_name, layer_type="field", overwrite=True
    )
    if (restart_stage > 2) and (stitch_size > 1):
        stitch_corrected_field = dst_stack.create_sublayer(
            stitch_corrected_name, layer_type="field", overwrite=False
        )
    if restart_stage <= 3 and compose_fields:
        corgie_logger.debug("Stitching blocks...")
        for block, stitch_block in zip(blocks[1:], stitch_blocks):
            block_bcube = block.broadcastable().get_bcube(bcube)
            block_list = block.get_neighbors(dist=decay_dist)
            corgie_logger.debug(f"src_block: {block}")
            corgie_logger.debug(f"influencing blocks: {block_list}")
            z_list = [b.stop for b in block_list]
            # stitch_corrected_field used if there is multi-section block overlap,
            # which requires voting to produce a corrected field.
            # If there is only single-section block overlap, then use
            # stitch_estimated_fields from each stitch_block
            if stitch_corrected_field is not None:
                stitching_fields = [stitch_corrected_field]
            else:
                # Order with furthest block first (convention of FieldSet).
                stitching_fields = [
                    stitch_block.dst_stack[stitch_estimated_name],
                    stitch_block.src_stack[stitch_estimated_name],
                ]

            broadcast_job = BroadcastJob(
                block_field=block.dst_stack[block_field_name],
                stitching_fields=stitching_fields,
                output_field=composed_field,
                chunk_xy=chunk_xy,
                bcube=block_bcube,
                pad=pad,
                z_list=z_list,
                mip=processor_mip[-1],
                decay_dist=decay_dist,
                blur_rate=blur_rate,
            )
            scheduler.register_job(
                broadcast_job, job_name=f"Broadcast {block} {block_bcube}"
            )

        scheduler.execute_until_completion()
        corgie_logger.debug("Done!")

        if len(blocks) > 1:
            block_bcube = blocks[0].get_bcube(bcube)
            copy_job = CopyLayerJob(
                src_layer=even_stack[block_field_name],
                dst_layer=composed_field,
                mip=processor_mip[-1],
                bcube=block_bcube,
                chunk_xy=chunk_xy,
                chunk_z=1,
            )
            scheduler.register_job(
                copy_job, job_name=f"Copy first block_field to composed_field location"
            )
            scheduler.execute_until_completion()
            corgie_logger.debug("Done!")

    if restart_stage <= 4 and final_render:
        if len(blocks) == 1:
            block_bcube = blocks[0].get_bcube(bcube)
            render_job = RenderJob(
                src_stack=src_stack,
                dst_stack=dst_stack,
                mips=processor_mip[-1],
                pad=pad,
                bcube=block_bcube,
                chunk_xy=chunk_xy,
                chunk_z=1,
                render_masks=True,
                blackout_masks=False,
                additional_fields=[even_stack[block_field_name]],
            )
            scheduler.register_job(
                render_job, job_name=f"Render first block {block_bcube}"
            )
        else:
            block_bcube = bcube.reset_coords(
                zs=blocks[0].start, ze=blocks[-1].stop, in_place=False
            )
            render_job = RenderJob(
                src_stack=src_stack,
                dst_stack=dst_stack,
                mips=processor_mip[-1],
                pad=pad,
                bcube=block_bcube,
                chunk_xy=chunk_xy,
                chunk_z=1,
                render_masks=True,
                blackout_masks=True,
                additional_fields=[composed_field],
            )
            scheduler.register_job(
                render_job, job_name=f"Render all blocks {block_bcube}"
            )
        scheduler.execute_until_completion()
        corgie_logger.debug("Done!")

    result_report = (
        f"Aligned layers {[str(l) for l in src_stack.get_layers_of_type('img')]}. "
        f"Results in {[str(l) for l in dst_stack.get_layers_of_type('img')]}"
    )
    corgie_logger.info(result_report)



import os
import click

from corgie.log import logger as corgie_logger
from corgie.boundingcube import get_bcube_from_coords
from corgie.cli.compute_stats import ComputeStatsJob
from corgie.argparsers import (
    LAYER_HELP_STR,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)
from corgie.cli.normalize import NormalizeJob
import json


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--spec_path",
    nargs=1,
    type=str,
    required=True,
    help="JSON spec relating src stacks, src z to dst z",
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)
@corgie_option("--suffix", "-s", nargs=1, type=str, default=None)
@corgie_optgroup("Normalize parameters")
@corgie_option("--recompute_stats/--no_recompute_stats", default=True)
@corgie_option("--stats_mip", "-m", nargs=1, type=int, required=None)
@corgie_option("--mip_start", "-m", nargs=1, type=int, required=True)
@corgie_option("--mip_end", "-e", nargs=1, type=int, required=True)
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=2048)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option("--mask_value", nargs=1, type=float, default=0.0)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def normalize_by_spec(
    ctx,
    src_layer_spec,
    spec_path,
    dst_folder,
    stats_mip,
    mip_start,
    mip_end,
    chunk_xy,
    chunk_z,
    start_coord,
    end_coord,
    coord_mip,
    suffix,
    recompute_stats,
    mask_value,
):
    if chunk_z != 1:
        raise NotImplemented(
            "Compute Statistics command currently only \
                supports per-section statistics."
        )
    result_report = ""
    scheduler = ctx.obj["scheduler"]

    if suffix is None:
        suffix = "_norm"
    else:
        suffix = f"_{suffix}"

    if stats_mip is None:
        stats_mip = mip_end

    with open(spec_path, "r") as f:
        spec = set(json.load(f))

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    img_layers = src_stack.get_layers_of_type("img")
    mask_layers = src_stack.get_layers_of_type("mask")
    field_layers = src_stack.get_layers_of_type("field")
    assert len(field_layers) == 0

    for l in img_layers:
        mean_layer = l.get_sublayer(
            name=f"mean{suffix}",
            path=os.path.join(dst_folder, f"mean{suffix}"),
            layer_type="section_value",
        )

        var_layer = l.get_sublayer(
            name=f"var{suffix}",
            path=os.path.join(dst_folder, f"var{suffix}"),
            layer_type="section_value",
        )

        if recompute_stats:
            for z in range(*bcube.z_range()):
                if z in spec:
                    job_bcube = bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
                    compute_stats_job = ComputeStatsJob(
                        src_layer=l,
                        mask_layers=mask_layers,
                        mean_layer=mean_layer,
                        var_layer=var_layer,
                        bcube=job_bcube,
                        mip=stats_mip,
                        chunk_xy=chunk_xy,
                        chunk_z=chunk_z,
                    )

                    # create scheduler and execute the job
                    scheduler.register_job(
                        compute_stats_job,
                        job_name=f"Compute Stats. Layer: {l}, Bcube: {job_bcube}",
                    )
            scheduler.execute_until_completion()

        dst_layer = l.get_sublayer(
            name=f"{l.name}{suffix}",
            path=os.path.join(dst_folder, "img", f"{l.name}{suffix}"),
            layer_type=l.get_layer_type(),
            dtype="float32",
            overwrite=True,
        )

        for z in range(*bcube.z_range()):
            if z in spec:
                job_bcube = bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
                result_report += f"Normalized {l} -> {dst_layer}\n"
                for mip in range(mip_start, mip_end + 1):
                    normalize_job = NormalizeJob(
                        src_layer=l,
                        mask_layers=mask_layers,
                        dst_layer=dst_layer,
                        mean_layer=mean_layer,
                        var_layer=var_layer,
                        stats_mip=stats_mip,
                        mip=mip,
                        bcube=job_bcube,
                        chunk_xy=chunk_xy,
                        chunk_z=chunk_z,
                        mask_value=mask_value,
                    )

                    # create scheduler and execute the job
                    scheduler.register_job(
                        normalize_job, job_name=f"Normalize {job_bcube}, MIP {mip}"
                    )
    scheduler.execute_until_completion()
    corgie_logger.info(result_report)



import click
import torch

from corgie import scheduling
from corgie.log import logger as corgie_logger
from corgie.data_backends import pass_data_backend
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie import argparsers
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
)


class DownsampleJob(scheduling.Job):
    def __init__(
        self,
        src_layer,
        mip_start,
        mip_end,
        bcube,
        chunk_xy,
        chunk_z,
        mips_per_task=3,
        dst_layer=None,
        preserve_zeros=False
    ):
        if dst_layer is None:
            dst_layer = src_layer
        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.mip_start = mip_start
        self.mip_end = mip_end
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.mips_per_task = mips_per_task
        self.preserve_zeros = preserve_zeros

        super().__init__()

    def task_generator(self):
        for mip in range(self.mip_start, self.mip_end, self.mips_per_task):
            this_mip_start = mip
            this_mip_end = min(self.mip_end, mip + self.mips_per_task)
            chunks = self.dst_layer.break_bcube_into_chunks(
                bcube=self.bcube,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip=this_mip_end,
                return_generator=True,
            )

            tasks = (
                DownsampleTask(
                    self.src_layer,
                    self.dst_layer,
                    this_mip_start,
                    this_mip_end,
                    input_chunk,
                    preserve_zeros=self.preserve_zeros
                )
                for input_chunk in chunks
            )
            print(
                "Yielding downsample tasks for bcube: {}, MIPs: {}-{}".format(
                    self.bcube, this_mip_start, this_mip_end
                )
            )

            yield tasks

            if mip == self.mip_start:
                self.src_layer = self.dst_layer

            # if not the last iteration
            if mip + self.mips_per_task < self.mip_end:
                yield scheduling.wait_until_done


class DownsampleTask(scheduling.Task):
    def __init__(self, src_layer, dst_layer, mip_start, mip_end, bcube, preserve_zeros):
        super().__init__()
        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.mip_start = mip_start
        self.mip_end = mip_end
        self.bcube = bcube
        self.preserve_zeros = preserve_zeros

    def execute(self):
        src_data = self.src_layer.read(bcube=self.bcube, mip=self.mip_start)
        # How to downsample depends on layer type.
        # Images are avg pooled, masks are max pooled, segmentation is...
        downsampler = self.src_layer.get_downsampler()
        for mip in range(self.mip_start, self.mip_end):
            dst_data = downsampler(src_data)
            if self.preserve_zeros:
                zeros = src_data == 0
                zeros_d = torch.nn.functional.max_pool2d(zeros.float(), 2, 2)
                dst_data[zeros_d > 0] = 0
            self.dst_layer.write(dst_data, bcube=self.bcube, mip=mip + 1)
            src_data = dst_data


@click.command()
@corgie_optgroup("Layer Parameterd")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    help="Specification for the source layer. " + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=False,
    help="Specification for the destination layer. "
    + "Refer to 'src_layer_spec' for parameter format."
    + " DEFAULT: Same as src_layer_spec",
)
@corgie_optgroup("Downsample parameters")
@corgie_option("--mip_start", "-m", nargs=1, type=int, required=True)
@corgie_option("--mip_end", "-e", nargs=1, type=int, required=True)
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=2048)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option("--mips_per_task", nargs=1, type=int, default=3)
@corgie_option("--preserve_zeros/--no_preserve_zeros", default=False)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def downsample(
    ctx,
    src_layer_spec,
    dst_layer_spec,
    mip_start,
    mip_end,
    chunk_xy,
    chunk_z,
    mips_per_task,
    start_coord,
    end_coord,
    coord_mip,
    preserve_zeros
):
    scheduler = ctx.obj["scheduler"]
    corgie_logger.debug("Setting up Source and Destination layers...")

    src_layer = create_layer_from_spec(
        src_layer_spec, caller_name="src layer", readonly=True
    )

    if dst_layer_spec is None:
        corgie_logger.info(
            "Destination layer not specified. Using Source layer " "as Destination."
        )
        dst_layer = src_layer
        dst_layer.readonly = False
    else:
        dst_layer = create_layer_from_spec(
            dst_layer_spec,
            caller_name="dst_layer layer",
            readonly=False,
            reference=src_layer,
            overwrite=True,
        )
    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)
    downsample_job = DownsampleJob(
        src_layer=src_layer,
        dst_layer=dst_layer,
        mip_start=mip_start,
        mip_end=mip_end,
        bcube=bcube,
        chunk_xy=chunk_xy,
        chunk_z=chunk_z,
        mips_per_task=mips_per_task,
        preserve_zeros=preserve_zeros,
    )

    # create scheduler and execute the job
    scheduler.register_job(downsample_job, job_name="downsample")
    scheduler.execute_until_completion()
    result_report = (
        f"Downsampled {src_layer} from {mip_start} to {mip_end}. Result in {dst_layer}"
    )
    corgie_logger.info(result_report)



import click

from corgie import scheduling, helpers, stack
from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)
import json


class MergeCopyJob(scheduling.Job):
    def __init__(self, src_stack, dst_stack, mip, bcube, z_list, chunk_xy):
        """Copy multiple images to the same destination image

        Args:
            src_stack (Stack)
            dst_stack (Stack)
            mip (int)
            bcube (BoundingCube)
            z_list ([int]): : ranked by layer priority (first image overwrites later images)
        """

        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.mip = mip
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.z_list = z_list

        super().__init__()

    def task_generator(self):
        chunks = self.dst_stack.get_layers()[0].break_bcube_into_chunks(
            bcube=self.bcube, chunk_xy=self.chunk_xy, chunk_z=1, mip=self.mip
        )

        tasks = [
            MergeCopyTask(
                src_stack=self.src_stack,
                dst_stack=self.dst_stack,
                mip=self.mip,
                bcube=input_chunk,
                z_list=self.z_list,
            )
            for input_chunk in chunks
        ]
        corgie_logger.info(
            f"Yielding copy tasks for bcube: {self.bcube}, MIP: {self.mip}"
        )

        yield tasks


class MergeCopyTask(scheduling.Task):
    def __init__(self, src_stack, dst_stack, mip, bcube, z_list):
        """Copy multiple images to the same destination image

        Args:
            src_stack (Stack)
            dst_stack (Stack)
            mip (int)
            bcube (BoundingCube)
            z_list ([int]): : ranked by layer priority (first image overwrites later images)
        """
        super().__init__(self)
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.mip = mip
        self.bcube = bcube
        self.z_list = z_list

    def execute(self):
        for k, z in enumerate(self.z_list[::-1]):
            src_bcube = self.bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            src_trans, src_data_dict = self.src_stack.read_data_dict(
                src_bcube, mip=self.mip, translation_adjuster=None, stack_name="src"
            )
            img_name = self.src_stack.get_layers_of_type("img")[0].name
            src_image = src_data_dict[f"src_{img_name}"]
            mask_name = self.src_stack.get_layers_of_type("mask")[0].name
            mask = src_data_dict[f"src_{mask_name}"]
            # mask_layers = src_stack.get_layers_of_type(["mask"])
            # mask = helpers.read_mask_list(mask_layers, src_bcube, self.mip)
            if k == 0:
                dst_image = src_image
                dst_image[~mask] = 0
            else:
                dst_image[mask] = src_image[mask]

        dst_layer = self.dst_stack.get_layers_of_type("img")[0]
        dst_layer.write(dst_image, bcube=self.bcube, mip=self.mip)


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Order img, mask, img, mask, etc. List must have length of multiple two."
    + LAYER_HELP_STR,
)
#
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=False,
    help="Destination folder for the copied stack",
)
@corgie_option(
    "--spec_path",
    nargs=1,
    type=str,
    required=True,
    help="JSON spec relating src stacks, src z to dst z",
)
@corgie_optgroup("Copy Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--mip", nargs=1, type=int, required=True)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@click.pass_context
def merge_copy(
    ctx,
    src_layer_spec,
    dst_folder,
    spec_path,
    chunk_xy,
    start_coord,
    end_coord,
    coord_mip,
    mip,
    suffix,
):

    scheduler = ctx.obj["scheduler"]
    if suffix is None:
        suffix = ""
    else:
        suffix = f"_{suffix}"

    corgie_logger.debug("Setting up layers...")
    assert len(src_layer_spec) % 2 == 0
    src_stacks = {}
    for k in range(len(src_layer_spec) // 2):
        src_stack = create_stack_from_spec(
            src_layer_spec[2 * k : 2 * k + 2], name="src", readonly=True
        )
        name = src_stack.get_layers_of_type("img")[0].path
        src_stacks[name] = src_stack

    with open(spec_path, "r") as f:
        spec = json.load(f)

    # if force_chunk_xy:
    #     force_chunk_xy = chunk_xy
    # else:
    #     force_chunk_xy = None

    # if force_chunk_z:
    #     force_chunk_z = chunk_z
    # else:
    #     force_chunk_z = None

    dst_stack = stack.create_stack_from_reference(
        reference_stack=list(src_stacks.values())[0],
        folder=dst_folder,
        name="dst",
        types=["img"],
        readonly=False,
        suffix=suffix,
        force_chunk_xy=None,
        force_chunk_z=None,
        overwrite=True,
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    for z in range(*bcube.z_range()):
        spec_z = str(z)
        if spec_z in spec.keys():
            src_dict = spec[str(z)]
            job_bcube = bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            src_stack = src_stacks[src_dict["cv_path"]]
            z_list = src_dict["z_list"]
            copy_job = MergeCopyJob(
                src_stack=src_stack,
                dst_stack=dst_stack,
                mip=mip,
                bcube=job_bcube,
                chunk_xy=chunk_xy,
                z_list=z_list,
            )
            # create scheduler and execute the job
            scheduler.register_job(copy_job, job_name="MergeCopy {}".format(job_bcube))
    scheduler.execute_until_completion()



from copy import deepcopy


class Block:
    def __init__(self, start, stop, src_stack, dst_stack, previous=None):
        self.start = start
        self.stop = stop
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.previous = previous

    def __str__(self):
        return f"Block {self.start}->{self.stop}, src_stack={self.src_stack.name}, dst_stack={self.dst_stack.name}"

    def __repr__(self):
        return self.__str__()

    def overlap(self, stitch_size):
        """Get block representing overlap between current and previous block

        Args:
            stitch_size (int): number of sections to use in stitching

        Returns:
            new Block
        """
        if self.previous is None:
            return None
        return Block(
            start=self.previous.stop,
            stop=self.previous.stop + stitch_size,
            src_stack=self.dst_stack,
            dst_stack=self.previous.dst_stack,
        )

    def broadcastable(self):
        """Get block representing the sections that are not block aligned by previous block

        Returns:
            new Block
        """
        if self.previous is None:
            return None
        return Block(
            start=self.previous.stop,
            stop=self.stop,
            src_stack=self.src_stack,
            dst_stack=self.dst_stack,
        )

    def get_neighbors(self, dist):
        """Get list of previous blocks where stop is within dist of current block's start.
        Order from furthest to nearest block (convention of FieldSet).

        Args:
            dist (int)

        Returns:
            list of blocks
        """
        neighbors = []
        prev = self.previous
        if prev is None:
            return neighbors
        while self.start - prev.stop < dist:
            neighbors.insert(0, prev)
            prev = prev.previous
            if prev is None:
                break
        return neighbors

    def get_bcube(self, bcube):
        return bcube.reset_coords(zs=self.start, ze=self.stop, in_place=False)


def get_blocks(
    start,
    stop,
    block_size,
    block_overlap,
    skip_list,
    src_stack,
    even_stack,
    odd_stack,
):
    partitions = partition(
        range(start, stop), sz=block_size, overlap=block_overlap, skip=skip_list
    )
    blocks = []
    previous = None
    for k, p in enumerate(partitions):
        dst_stack = even_stack if k % 2 == 0 else odd_stack
        block = Block(
            start=p.start,
            stop=p.stop,
            src_stack=src_stack,
            dst_stack=dst_stack,
            previous=previous,
        )
        blocks.append(block)
        previous = block
    return blocks


def partition(xrange, sz, overlap=0, skip=[]):
    """Divide range into subsets specified by size, overlap, and indices to skip.

    Args:
        xrange (range)
        sz (int): target size of subset
        overlap (int): amount of overlap between subsets
        skip ([int]): indices to skip in start/end of subset

    Returns:
        list of subset ranges
    """
    subsets = []
    xs_start = xrange.start
    xs_next = min(xs_start + sz, xrange.stop)
    while xs_start < xs_next:
        xs_stop = xs_next
        while xs_stop in skip and xs_stop < xrange.stop:
            xs_stop += 1
        while xs_start in skip and xs_start + overlap < xs_stop:
            xs_start += 1
        xs_stop = min(xrange.stop, xs_stop + overlap)
        if xs_start < xs_stop:
            subsets.append(range(xs_start, xs_stop))
        xs_start = xs_next
        xs_next = min(xs_next + sz, xrange.stop)
    return subsets



import click

from corgie import scheduling, helpers, stack
from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

class CopyLayerJob(scheduling.Job):
    def __init__(
        self,
        src_layer,
        dst_layer,
        mip,
        bcube,
        chunk_xy,
        chunk_z,
    ):
        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.mip = mip
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z

        super().__init__()

    def task_generator(self):
        chunks = self.dst_layer.break_bcube_into_chunks(
            bcube=self.bcube,
            chunk_xy=self.chunk_xy,
            chunk_z=self.chunk_z,
            mip=self.mip,
            return_generator=True,
        )

        tasks = (
            CopyLayerTask(
                self.src_layer,
                self.dst_layer,
                mip=self.mip,
                bcube=input_chunk,
            )
            for input_chunk in chunks
        )
        corgie_logger.info(
            f"Yielding copy layer tasks for bcube: {self.bcube}, MIP: {self.mip}"
        )

        yield tasks


class CopyLayerTask(scheduling.Task):
    def __init__(self, src_layer, dst_layer, mip, bcube):
        super().__init__(self)
        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.mip = mip
        self.bcube = bcube

    def execute(self):
        src = self.src_layer.read(bcube=self.bcube, mip=self.mip)
        self.dst_layer.write(src, bcube=self.bcube, mip=self.mip)


class CopyJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        dst_stack,
        mip,
        copy_masks,
        blackout_masks,
        bcube,
        chunk_xy,
        chunk_z,
    ):
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.mip = mip
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.copy_masks = copy_masks
        self.blackout_masks = blackout_masks

        super().__init__()

    def task_generator(self):
        chunks = self.dst_stack.get_layers()[0].break_bcube_into_chunks(
            bcube=self.bcube,
            chunk_xy=self.chunk_xy,
            chunk_z=self.chunk_z,
            mip=self.mip,
            return_generator=True,
        )

        tasks = (
            CopyTask(
                self.src_stack,
                self.dst_stack,
                blackout_masks=self.blackout_masks,
                copy_masks=self.copy_masks,
                mip=self.mip,
                bcube=input_chunk,
            )
            for input_chunk in chunks
        )
        corgie_logger.info(
            f"Yielding copy tasks for bcube: {self.bcube}, MIP: {self.mip}"
        )

        yield tasks


class CopyTask(scheduling.Task):
    def __init__(self, src_stack, dst_stack, copy_masks, blackout_masks, mip, bcube):
        super().__init__(self)
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.copy_masks = copy_masks
        self.blackout_masks = blackout_masks
        self.mip = mip
        self.bcube = bcube

    def execute(self):
        src_translation, src_data_dict = self.src_stack.read_data_dict(
            self.bcube, mip=self.mip, translation_adjuster=None, stack_name="src"
        )

        if self.blackout_masks:
            mask_layers = self.dst_stack.get_layers_of_type(["mask"])
            mask = helpers.read_mask_list(mask_layers, self.bcube, self.mip)
        else:
            mask = None
        if self.copy_masks:
            write_layers = self.dst_stack.get_layers_of_type(["img", "mask"])
        else:
            write_layers = self.dst_stack.get_layers_of_type("img")

        for l in write_layers:
            src = src_data_dict[f"src_{l.name}"]
            if mask is not None:
                src[mask] = 0
            l.write(src, bcube=self.bcube, mip=self.mip)

        # copy fields
        write_layers = self.dst_stack.get_layers_of_type("field")
        for l in write_layers:
            src = src_data_dict[f"src_{l.name}"]
            l.write(src, bcube=self.bcube, mip=self.mip)


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
#
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=False,
    help="Destination folder for the copied stack",
)
@corgie_optgroup("Copy Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option("--mip", nargs=1, type=int, required=True)
@corgie_option("--blackout_masks/--no_blackout_masks", default=False)
@corgie_option("--copy_masks/--no_copy_masks", default=True)
@corgie_option("--force_chunk_xy", nargs=1, type=int,
   help="Will force the chunking of the underlying cloudvolume"
)
@corgie_option("--force_chunk_z", nargs=1, type=int,
   help="Will force the chunking of the underlying cloudvolume"
)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@click.pass_context
def copy(
    ctx,
    src_layer_spec,
    dst_folder,
    copy_masks,
    blackout_masks,
    chunk_xy,
    chunk_z,
    start_coord,
    end_coord,
    coord_mip,
    mip,
    suffix,
    force_chunk_xy,
    force_chunk_z,
):

    scheduler = ctx.obj["scheduler"]
    if suffix is None:
        suffix = ""
    else:
        suffix = f"_{suffix}"

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    if not force_chunk_xy:
        force_chunk_xy = chunk_xy

    if not force_chunk_z:
        force_chunk_z = chunk_z

    dst_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=dst_folder,
        name="dst",
        types=["img", "mask"],
        readonly=False,
        suffix=suffix,
        force_chunk_xy=force_chunk_xy,
        force_chunk_z=force_chunk_z,
        overwrite=True,
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    copy_job = CopyJob(
        src_stack=src_stack,
        dst_stack=dst_stack,
        mip=mip,
        bcube=bcube,
        chunk_xy=chunk_xy,
        chunk_z=chunk_z,
        copy_masks=copy_masks,
        blackout_masks=blackout_masks,
    )
    # create scheduler and execute the job
    scheduler.register_job(copy_job, job_name="Copy {}".format(bcube))
    scheduler.execute_until_completion()

    result_report = (
        f"Copied layers {[str(l) for l in src_stack.get_layers_of_type('img')]}. "
        f"to {[str(l) for l in dst_stack.get_layers_of_type('img')]}"
    )
    corgie_logger.info(result_report)



import os
import click

from corgie import scheduling
from corgie import helpers, stack
from corgie.log import logger as corgie_logger
from corgie.boundingcube import get_bcube_from_coords
from corgie.cli.compute_stats import ComputeStatsJob
from corgie.argparsers import (
    LAYER_HELP_STR,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)


class NormalizeJob(scheduling.Job):
    def __init__(
        self,
        src_layer,
        mask_layers,
        dst_layer,
        mean_layer,
        var_layer,
        stats_mip,
        mip,
        bcube,
        chunk_xy,
        chunk_z,
        mask_value,
    ):
        self.src_layer = src_layer
        self.mask_layers = mask_layers
        self.dst_layer = dst_layer
        self.var_layer = var_layer
        self.mean_layer = mean_layer
        self.stats_mip = stats_mip
        self.mip = mip
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.mask_value = mask_value
        super().__init__()

    def task_generator(self):
        chunks = self.dst_layer.break_bcube_into_chunks(
            bcube=self.bcube,
            chunk_xy=self.chunk_xy,
            chunk_z=self.chunk_z,
            mip=self.mip,
            return_generator=True,
        )

        tasks = (
            NormalizeTask(
                self.src_layer,
                self.mask_layers,
                self.dst_layer,
                self.mean_layer,
                self.var_layer,
                self.stats_mip,
                self.mip,
                self.mask_value,
                input_chunk,
            )
            for input_chunk in chunks
        )
        print(f"Yielding normalize tasks for bcube: {self.bcube}, MIP: {self.mip}")

        yield tasks


class NormalizeTask(scheduling.Task):
    def __init__(
        self,
        src_layer,
        mask_layers,
        dst_layer,
        mean_layer,
        var_layer,
        stats_mip,
        mip,
        mask_value,
        bcube,
    ):
        super().__init__()
        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.mean_layer = mean_layer
        self.mask_layers = mask_layers
        self.var_layer = var_layer
        self.stats_mip = stats_mip
        self.mip = mip
        self.bcube = bcube
        self.mask_value = mask_value

    def execute(self):
        corgie_logger.info(
            f"Normalizing {self.src_layer} at MIP{self.mip}, region: {self.bcube}"
        )
        mean_data = self.mean_layer.read(self.bcube, mip=self.stats_mip)
        var_data = self.var_layer.read(self.bcube, mip=self.stats_mip)

        src_data = self.src_layer.read(self.bcube, mip=self.mip)
        mask_data = helpers.read_mask_list(
            mask_list=self.mask_layers, bcube=self.bcube, mip=self.mip
        )

        dst_data = (src_data - mean_data) / var_data.sqrt()
        if mask_data is not None:
            dst_data[mask_data] = self.mask_value
        self.dst_layer.write(dst_data, self.bcube, mip=self.mip)


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)
@corgie_option("--suffix", "-s", nargs=1, type=str, default=None)
@corgie_optgroup("Normalize parameters")
@corgie_option("--recompute_stats/--no_recompute_stats", default=True)
@corgie_option("--stats_mip", "-m", nargs=1, type=int, required=None)
@corgie_option("--mip_start", "-m", nargs=1, type=int, required=True)
@corgie_option("--mip_end", "-e", nargs=1, type=int, required=True)
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=2048)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option("--force_chunk_xy", nargs=1, type=int)
@corgie_option("--mask_value", nargs=1, type=float, default=0.0)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def normalize(
    ctx,
    src_layer_spec,
    dst_folder,
    stats_mip,
    mip_start,
    mip_end,
    chunk_xy,
    chunk_z,
    force_chunk_xy,
    start_coord,
    end_coord,
    coord_mip,
    suffix,
    recompute_stats,
    mask_value,
):
    if chunk_z != 1:
        raise NotImplemented(
            "Compute Statistics command currently only \
                supports per-section statistics."
        )
    result_report = ""
    scheduler = ctx.obj["scheduler"]

    if suffix is None:
        suffix = "_norm"
    else:
        suffix = f"_{suffix}"

    if stats_mip is None:
        stats_mip = mip_end

    if not force_chunk_xy:
        force_chunk_xy = chunk_xy

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    dst_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=dst_folder,
        name="dst",
        types=["img"],
        readonly=False,
        suffix=suffix,
        overwrite=True,
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    img_layers = src_stack.get_layers_of_type("img")
    mask_layers = src_stack.get_layers_of_type("mask")
    field_layers = src_stack.get_layers_of_type("field")
    assert len(field_layers) == 0

    for l in img_layers:
        mean_layer = l.get_sublayer(
            name=f"mean_{l.name}{suffix}",
            path=os.path.join(dst_folder, f"mean_{l.name}{suffix}"),
            layer_type="section_value",
        )

        var_layer = l.get_sublayer(
            name=f"var_{l.name}{suffix}",
            path=os.path.join(dst_folder, f"var_{l.name}{suffix}"),
            layer_type="section_value",
        )

        if recompute_stats:
            compute_stats_job = ComputeStatsJob(
                src_layer=l,
                mask_layers=mask_layers,
                mean_layer=mean_layer,
                var_layer=var_layer,
                bcube=bcube,
                mip=stats_mip,
                chunk_xy=chunk_xy,
                chunk_z=chunk_z,
            )

            # create scheduler and execute the job
            scheduler.register_job(
                compute_stats_job, job_name=f"Compute Stats. Layer: {l}, Bcube: {bcube}"
            )
            scheduler.execute_until_completion()

        dst_layer = l.get_sublayer(
            name=f"{l.name}{suffix}",
            path=os.path.join(dst_folder, "img", f"{l.name}{suffix}"),
            layer_type=l.get_layer_type(),
            dtype="float32",
            force_chunk_xy=force_chunk_xy,
            overwrite=True,
        )

        result_report += f"Normalized {l} -> {dst_layer}\n"
        for mip in range(mip_start, mip_end + 1):
            normalize_job = NormalizeJob(
                src_layer=l,
                mask_layers=mask_layers,
                dst_layer=dst_layer,
                mean_layer=mean_layer,
                var_layer=var_layer,
                stats_mip=stats_mip,
                mip=mip,
                bcube=bcube,
                chunk_xy=chunk_xy,
                chunk_z=chunk_z,
                mask_value=mask_value,
            )

            # create scheduler and execute the job
            scheduler.register_job(
                normalize_job, job_name=f"Normalize {bcube}, MIP {mip}"
            )
    scheduler.execute_until_completion()
    corgie_logger.info(result_report)



import click
import procspec
import torchfields

from corgie import scheduling, argparsers, helpers

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

from corgie.cli.common import ChunkedJob


class InvertFieldJob(scheduling.Job):
    def __init__(
        self,
        src_layer,
        dst_layer,
        chunk_xy,
        chunk_z,
        mip,
        pad,
        bcube,
        blend_xy=0,
        crop=None,
    ):

        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.blend_xy = blend_xy
        self.pad = pad
        self.crop = pad
        if crop is not None:
            self.crop = crop
        self.bcube = bcube
        self.mip = mip

        super().__init__()

    def task_generator(self):
        this_task = helpers.PartialSpecification(
            InvertFieldTask,
            src_layer=self.src_layer,
            pad=self.pad,
            crop=self.crop,
        )

        chunked_job = ChunkedJob(
            task_class=this_task,
            dst_layer=self.dst_layer,
            chunk_xy=self.chunk_xy,
            chunk_z=self.chunk_z,
            blend_xy=self.blend_xy,
            mip=self.mip,
            bcube=self.bcube,
        )

        yield from chunked_job.task_generator


class InvertFieldTask(scheduling.Task):
    def __init__(self, src_layer, dst_layer, mip, pad, crop, bcube):
        super().__init__()
        self.src_layer = src_layer
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.crop = crop
        self.bcube = bcube

    def execute(self):
        src_bcube = self.bcube.uncrop(self.pad, self.mip)

        src_data = self.src_layer.read(src_bcube, mip=self.mip)

        result = (~src_data.field().from_pixels()).pixels()

        cropped_result = helpers.crop(result, self.crop)

        self.dst_layer.write(cropped_result, bcube=self.bcube, mip=self.mip)


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=False,
    help="Source layer spec. " + LAYER_HELP_STR,
)
#
@corgie_option(
    "--dst_layer_spec",
    nargs=1,
    type=str,
    required=True,
    help="Specification for the destination layer. Must be a field type.",
)
@corgie_optgroup("Invert Field Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option("--blend_xy", nargs=1, type=int, default=0)
@corgie_option("--pad", nargs=1, type=int, default=128)
@corgie_option("--crop", nargs=1, type=int, default=None)
@corgie_option("--mip", nargs=1, type=int)
@corgie_option("--force_chunk_xy", is_flag=True)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def invert_field(
    ctx,
    src_layer_spec,
    dst_layer_spec,
    pad,
    crop,
    chunk_xy,
    start_coord,
    mip,
    end_coord,
    coord_mip,
    blend_xy,
    chunk_z,
    force_chunk_xy,
):
    scheduler = ctx.obj["scheduler"]

    if force_chunk_xy:
        force_chunk_xy = chunk_xy
    else:
        force_chunk_xy = None

    corgie_logger.debug("Setting up layers...")
    src_layer = create_layer_from_spec(
        src_layer_spec,
        allowed_types=["field"],
        default_type="field",
        readonly=True,
        caller_name="src_layer",
    )

    dst_layer = create_layer_from_spec(
        dst_layer_spec,
        allowed_types=["field"],
        default_type="field",
        readonly=False,
        caller_name="dst_layer",
        reference=src_layer,
        overwrite=True,
        force_chunk_xy=force_chunk_xy,
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    if crop is None:
        crop = pad

    invert_field_job = InvertFieldJob(
        src_layer=src_layer,
        dst_layer=dst_layer,
        chunk_xy=chunk_xy,
        chunk_z=chunk_z,
        blend_xy=blend_xy,
        pad=pad,
        mip=mip,
        crop=crop,
        bcube=bcube,
    )

    # create scheduler and execute the job
    scheduler.register_job(invert_field_job, job_name="Invert Field {}".format(bcube))
    scheduler.execute_until_completion()



from copy import deepcopy
from math import log
from corgie.log import logger as corgie_logger
from corgie import scheduling, helpers
from corgie.stack import PyramidDistanceFieldSet


class BroadcastJob(scheduling.Job):
    def __init__(
        self,
        block_field,
        stitching_fields,
        output_field,
        chunk_xy,
        bcube,
        pad,
        z_list,
        mip,
        decay_dist,
        blur_rate,
    ):
        """
        Args:
            block_field (Layer): final field
            stitching_fields ([Layers]): collection of fields that will warp in order
            output_field (Layer)
            mip (int)
            bcube (BoundingCube)
            pad (int)
            z_list ([int]): list of z indices where stitching_fields should be sampled
            decay_dist (float): distance for influence of previous section
            blur_rate (float)
        """
        self.block_field = block_field
        self.stitching_fields = stitching_fields
        self.output_field = output_field
        self.chunk_xy = chunk_xy
        self.bcube = bcube
        self.pad = pad
        self.z_list = z_list
        self.mip = mip
        self.decay_dist = decay_dist
        self.blur_rate = blur_rate
        super().__init__()

    def task_generator(self):
        chunks = self.output_field.break_bcube_into_chunks(
            bcube=self.bcube, chunk_xy=self.chunk_xy, chunk_z=1, mip=self.mip
        )

        tasks = []
        for chunk in chunks:
            task = BroadcastTask(
                block_field=self.block_field,
                stitching_fields=self.stitching_fields,
                output_field=self.output_field,
                mip=self.mip,
                bcube=chunk,
                pad=self.pad,
                z_list=self.z_list,
                decay_dist=self.decay_dist,
                blur_rate=self.blur_rate,
            )
            tasks.append(task)

        corgie_logger.debug(
            "Yielding BroadcastTask for bcube: {}, MIP: {}".format(self.bcube, self.mip)
        )
        yield tasks


class ComposeWithDistanceTask(scheduling.Task):
    def __init__(
        self, input_fields, output_field, mip, bcube, pad, z_list, decay_dist, blur_rate
    ):
        """Compose set of fields, adjusted by distance

        Order of fields are from target to source, e.g.
            $f_{0 \leftarrow 2} \circ f_{2 \leftarrow 3}$ : [0, 2, 3]

        Args:
            input_fields ([Layers]): collection of fields
            output_field (Layer)
            mip (int)
            bcube (BoundingCube)
            pad (int)
            z_list ([int]): list of z indices
            decay_dist (float): distance for influence of previous section
            blur_rate (float)
        """
        super().__init__()
        self.input_fields = input_fields
        self.output_field = output_field
        self.mip = mip
        self.pad = pad
        self.bcube = bcube
        self.z_list = z_list
        self.decay_dist = decay_dist
        self.blur_rate = blur_rate
        self.trans_adj = helpers.percentile_trans_adjuster

    def execute(self):
        corgie_logger.debug(f"ComposeWithDistanceTask, {self.bcube}")
        corgie_logger.debug(f"input_fields: {self.input_fields}")
        corgie_logger.debug(f"z_list: {self.z_list}")
        pbcube = self.bcube.uncrop(self.pad, self.mip)
        fields = PyramidDistanceFieldSet(
            decay_dist=self.decay_dist,
            blur_rate=self.blur_rate,
            layers=self.input_fields,
        )
        field = fields.read(bcube=pbcube, z_list=self.z_list, mip=self.mip)
        cropped_field = helpers.crop(field, self.pad)
        self.output_field.write(cropped_field, bcube=self.bcube, mip=self.mip)


class BroadcastTask(scheduling.Task):
    def __init__(
        self,
        block_field,
        stitching_fields,
        output_field,
        mip,
        bcube,
        pad,
        z_list,
        decay_dist,
        blur_rate,
    ):
        """Compose set of stitching_fields, adjusted by distance, with block_field.

        Args:
            block_field (Layer): most recent field, that last to be warped
            stitching_fields ([Layers]): collection of fields at stitching interfaces,
                which are assumed to alternate
            output_field (Layer)
            mip (int)
            bcube (BoundingCube): xy location of where to sample all fields;
                z location of where to sample block_field and where to write composed field
            pad (int)
            z_list ([int]): list of z locations for where to sample each stitching_field;
                length of z_list will indicate how many times to repeat the stitching_fields list
            decay_dist (float): distance for influence of previous section
            blur_rate (float)
        """
        super().__init__()
        self.block_field = block_field
        self.stitching_fields = stitching_fields
        self.output_field = output_field
        self.mip = mip
        self.pad = pad
        self.bcube = bcube
        self.z_list = z_list
        self.decay_dist = decay_dist
        self.blur_rate = blur_rate
        self.trans_adj = helpers.percentile_trans_adjuster

    def execute(self):
        corgie_logger.debug(f"BroadcastTask, {self.bcube}")
        input_fields = self.stitching_fields[::-1]
        if len(self.z_list) != len(input_fields):
            fmul = len(self.z_list) // len(input_fields)
            frem = len(self.z_list) % len(input_fields)
            input_fields = input_fields[::-1]
            input_fields = fmul * input_fields + input_fields[:frem]
            input_fields = input_fields[::-1]
        input_fields += [self.block_field]
        z_list = self.z_list + [self.bcube.z_range()[0]]
        corgie_logger.debug(f"input_fields: {input_fields}")
        corgie_logger.debug(f"z_list: {z_list}")
        pbcube = self.bcube.uncrop(self.pad, self.mip)
        corgie_logger.debug(f"pbcube: {pbcube}")
        fields = PyramidDistanceFieldSet(
            decay_dist=self.decay_dist, blur_rate=self.blur_rate, layers=input_fields
        )
        field = fields.read(bcube=pbcube, z_list=z_list, mip=self.mip)
        cropped_field = helpers.crop(field, self.pad)
        self.output_field.write(cropped_field, bcube=self.bcube, mip=self.mip)



import click
import torch
import procspec

from corgie import scheduling, helpers, stack

from corgie.log import logger as corgie_logger
from corgie.boundingcube import get_bcube_from_coords

from corgie.argparsers import (
    LAYER_HELP_STR,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

from corgie.cli.render import RenderJob
from corgie.cli.common import ChunkedJob


class SeethroughBlockJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        dst_stack,
        render_method,
        bcube,
        seethrough_method=None,
        suffix=None,
    ):
        """Write out a block sequentially using seethrough method

        This is useful when a model requires a filled in image.
        """
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.bcube = bcube
        self.seethrough_method = seethrough_method
        self.render_method = render_method
        self.suffix = suffix
        super().__init__()

    def task_generator(self):
        seethrough_offset = -1
        seethrough_mask_layer = self.dst_stack.create_sublayer(
            f"seethrough_mask{self.suffix}",
            layer_type="mask",
            overwrite=True,
            force_chunk_z=1,
        )
        z_start, z_stop = self.bcube.z_range()
        for z in range(z_start, z_stop):
            bcube = self.bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            if z == z_start:
                corgie_logger.debug(f"Copy section {z}")
                render_job = self.render_method(
                    src_stack=self.src_stack,
                    dst_stack=self.dst_stack,
                    bcube=bcube,
                    mips=self.seethrough_method.mip,
                    blackout_masks=True,
                )

                yield from render_job.task_generator
                yield scheduling.wait_until_done
            else:
                # Now, we'll apply misalignment detection to produce a mask
                # this mask will be used in the final render step
                seethrough_mask_job = self.seethrough_method(
                    src_stack=self.src_stack,
                    tgt_stack=self.dst_stack,
                    bcube=bcube,
                    tgt_z_offset=seethrough_offset,
                    suffix=self.suffix,
                    dst_layer=seethrough_mask_layer,
                )

                yield from seethrough_mask_job.task_generator
                yield scheduling.wait_until_done
                corgie_logger.debug(f"Render {z}")
                render_job = self.render_method(
                    src_stack=self.src_stack,
                    dst_stack=self.dst_stack,
                    bcube=bcube,
                    blackout_masks=False,
                    seethrough_mask_layer=seethrough_mask_layer,
                    seethrough_offset=seethrough_offset,
                    mips=self.seethrough_method.mip,
                )
                yield from render_job.task_generator
                yield scheduling.wait_until_done


class SeethroughCompareJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        dst_layer,
        chunk_xy,
        processor_spec,
        mip,
        pad,
        crop,
        bcube,
        tgt_z_offset,
        tgt_stack=None,
        suffix="",
        seethrough_limit=None,
        pixel_offset_layer=None,
    ):
        self.src_stack = src_stack
        if tgt_stack is None:
            tgt_stack = src_stack

        self.tgt_stack = tgt_stack
        self.dst_layer = dst_layer
        self.chunk_xy = chunk_xy
        self.pad = pad
        self.crop = crop
        self.bcube = bcube
        self.tgt_z_offset = tgt_z_offset

        self.suffix = suffix

        self.processor_spec = processor_spec
        self.mip = mip
        self.pixel_offset_layer = pixel_offset_layer
        if seethrough_limit is None or seethrough_limit == tuple():
            # If limit not specified, no limit
            self.seethrough_limit = [0] * len(self.processor_spec)
        else:
            self.seethrough_limit = seethrough_limit

        self._validate_seethrough()

        super().__init__()

    def task_generator(self):
        for i in range(len(self.processor_spec)):
            cs_task = helpers.PartialSpecification(
                SeethroughCompareTask,
                processor_spec=self.processor_spec[i],
                tgt_z_offset=self.tgt_z_offset,
                src_stack=self.src_stack,
                pad=self.pad,
                crop=self.crop,
                tgt_stack=self.tgt_stack,
                seethrough_limit=self.seethrough_limit[i],
                pixel_offset_layer=self.pixel_offset_layer,
            )

            chunked_job = ChunkedJob(
                task_class=cs_task,
                dst_layer=self.dst_layer,
                chunk_xy=self.chunk_xy,
                chunk_z=1,
                mip=self.mip,
                bcube=self.bcube,
                suffix=self.suffix,
            )

            yield from chunked_job.task_generator

            # Each seethrough processor writes to the same mask layer, so we
            # wait for each processor to finish to avoid race conditions.
            if i < len(self.processor_spec) - 1:
                yield scheduling.wait_until_done

    def _validate_seethrough(self):
        num_ps = len(self.processor_spec)
        num_sl = len(self.seethrough_limit)
        if num_ps != num_sl:
            raise ValueError(
                f"{num_ps} processors and {num_sl} seethrough limits specified to a SeethroughCompareJob. These must be equal."
            )
        for sl in self.seethrough_limit:
            if type(sl) != int:
                raise ValueError(f"Specified seethrough limit {sl} is not an integer")
            if sl < 0:
                raise ValueError(
                    f"Seethrough limits to SeethroughCompareJobs must be non-negative."
                )


class SeethroughCompareTask(scheduling.Task):
    def __init__(
        self,
        processor_spec,
        src_stack,
        tgt_stack,
        dst_layer,
        mip,
        pad,
        crop,
        tgt_z_offset,
        bcube,
        seethrough_limit,
        pixel_offset_layer,
    ):
        super().__init__()
        self.processor_spec = processor_spec
        self.src_stack = src_stack
        self.tgt_stack = tgt_stack
        self.dst_layer = dst_layer
        self.mip = mip
        self.pad = pad
        self.crop = crop
        self.tgt_z_offset = tgt_z_offset
        self.bcube = bcube
        self.seethrough_limit = seethrough_limit
        self.pixel_offset_layer = pixel_offset_layer

    def execute(self):
        src_bcube = self.bcube.uncrop(self.pad, self.mip)
        tgt_bcube = src_bcube.translate(z_offset=self.tgt_z_offset)

        processor = procspec.parse_proc(spec_str=self.processor_spec)

        _, tgt_data_dict = self.tgt_stack.read_data_dict(
            tgt_bcube, mip=self.mip, stack_name="tgt"
        )

        _, src_data_dict = self.src_stack.read_data_dict(
            src_bcube, mip=self.mip, stack_name="src"
        )

        processor_input = {**src_data_dict, **tgt_data_dict}

        result = processor(processor_input, output_key="result")

        tgt_pixel_data = self.pixel_offset_layer.read(
            bcube=self.bcube.translate(z_offset=self.tgt_z_offset), mip=self.mip
        )
        written_pixel_data = self.pixel_offset_layer.read(
            bcube=self.bcube, mip=self.mip
        )
        written_mask_data = self.dst_layer.read(bcube=self.bcube, mip=self.mip)
        result = result.to(device=written_mask_data.device)
        cropped_result = helpers.crop(result, self.crop)
        if self.seethrough_limit > 0:
            seethrough_mask = (cropped_result > 0) & (
                tgt_pixel_data < self.seethrough_limit
            )
        else:
            seethrough_mask = cropped_result > 0
        written_mask_data[seethrough_mask] = True
        written_pixel_data[seethrough_mask] = (
            torch.minimum(
                tgt_pixel_data[seethrough_mask],
                torch.ones_like(tgt_pixel_data[seethrough_mask]) * 254,
            )
            + 1
        )
        self.dst_layer.write(written_mask_data, bcube=self.bcube, mip=self.mip)
        self.pixel_offset_layer.write(
            written_pixel_data, bcube=self.bcube, mip=self.mip
        )


@click.command()
# Layers
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)
@corgie_option("--suffix", nargs=1, type=str, default=None)
@corgie_optgroup("Render Method Specification")
@corgie_option(
    "--seethrough_spec",
    nargs=1,
    type=str,
    default=None,
    multiple=True,
    help="Seethrough method spec. Use multiple times to specify different methods (e.g. seethrough misalignments, seethrough black, etc.)",
)
@corgie_option(
    "--seethrough_limit",
    nargs=1,
    type=int,
    default=None,
    multiple=True,
    help="For each seethrough method, how many sections are allowed to be seenthrough. 0 or None means no limit.",
)
@corgie_option("--chunk_xy", nargs=1, type=int, default=1024)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def seethrough_block(
    ctx,
    src_layer_spec,
    dst_folder,
    chunk_xy,
    start_coord,
    end_coord,
    coord_mip,
    suffix,
    seethrough_spec,
    seethrough_limit,
    seethrough_spec_mip,
    force_chunk_z=1,
):
    scheduler = ctx.obj["scheduler"]

    if suffix is None:
        suffix = "_seethrough"
    else:
        suffix = f"_{suffix}"

    crop, pad = 0, 0
    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)
    src_stack.folder = dst_folder
    dst_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=dst_folder,
        name="dst",
        types=["img", "mask"],
        readonly=False,
        suffix=suffix,
        overwrite=True,
        force_chunk_z=force_chunk_z,
    )
    render_method = helpers.PartialSpecification(
        f=RenderJob, pad=pad, chunk_xy=chunk_xy, chunk_z=1, render_masks=False,
    )
    seethrough_method = helpers.PartialSpecification(
        f=SeethroughCompareJob,
        mip=seethrough_spec_mip,
        processor_spec=seethrough_spec,
        chunk_xy=chunk_xy,
        pad=pad,
        crop=pad,
        seethrough_limit=seethrough_limit,
    )
    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    seethrough_block_job = SeethroughBlockJob(
        src_stack=src_stack,
        dst_stack=dst_stack,
        bcube=bcube,
        render_method=render_method,
        seethrough_method=seethrough_method,
        suffix=suffix,
    )
    # create scheduler and execute the job
    scheduler.register_job(
        seethrough_block_job, job_name="Seethrough Block {}".format(bcube)
    )

    scheduler.execute_until_completion()
    result_report = (
        f"Rendered layers {[str(l) for l in src_stack.get_layers_of_type('img')]}. "
        f"Results in {[str(l) for l in dst_stack.get_layers_of_type('img')]}"
    )
    corgie_logger.info(result_report)



import json
import six
import copy
import cachetools

import cloudvolume
from cloudvolume import CloudVolume, Storage

from corgie.log import logger as corgie_logger


def jsonize_key(*kargs, **kwargs):
    result = ""
    for k in kargs[1:]:
        result += json.dumps(k)
        result += "_"

    result += json.dumps(kwargs, sort_keys=True)
    return result


def cv_is_cached(*kargs, **kwargs):
    key = jsonize_key(*kargs, **kwargs)
    return key in cv_cache


cv_cache = cachetools.LRUCache(maxsize=500)


class CachedCloudVolume(CloudVolume):
    @cachetools.cached(cv_cache, key=jsonize_key)
    def __new__(self, *args, **kwargs):
        return super().__new__(self, *args, **kwargs)


def deserialize_miplessCV_old(s, cache={}):
    if s in cache:
        return cache[s]
    else:
        contents = json.loads(s)
        mcv = MiplessCloudVolume(
            contents["path"], mkdir=contents["mkdir"], **contents["kwargs"]
        )
        cache[s] = mcv
        return mcv


def deserialize_miplessCV_old2(s, cache={}):
    cv_kwargs = {
        "bounded": False,
        "progress": False,
        "autocrop": False,
        "non_aligned_writes": False,
        "cdn_cache": False,
    }
    if s in cache:
        return cache[s]
    else:
        contents = json.loads(s)
        mcv = MiplessCloudVolume(
            contents["path"], mkdir=False, fill_missing=True, **cv_kwargs
        )
        cache[s] = mcv
        return mcv


def deserialize_miplessCV(s, cache={}):
    cv_kwargs = {
        "bounded": False,
        "progress": False,
        "autocrop": False,
        "non_aligned_writes": False,
        "cdn_cache": False,
    }
    if s in cache:
        return cache[s]
    else:
        mcv = MiplessCloudVolume(s, mkdir=False, fill_missing=True, **cv_kwargs)
        cache[s] = mcv
        return mcv


class MiplessCloudVolume:
    """Multi-mip access to CloudVolumes using the same path"""

    def __init__(
        self,
        path,
        info=None,
        allow_info_writes=True,
        obj=CachedCloudVolume,
        default_chunk=(512, 512, 1),
        overwrite=False,
        **kwargs
    ):
        self.path = path
        self.allow_info_writes = allow_info_writes
        self.cv_params = {}
        self.cv_params["info"] = info
        if "cv_params" in kwargs:
            self.cv_params.update(kwargs["cv_params"])
        self.cv_params.setdefault("bounded", False)
        self.cv_params.setdefault("progress", False)
        self.cv_params.setdefault("autocrop", False)
        self.cv_params.setdefault("non_aligned_writes", False)
        self.cv_params.setdefault("cdn_cache", False)
        self.cv_params.setdefault("fill_missing", True)
        self.cv_params.setdefault("delete_black_uploads", True)
        self.cv_params.setdefault("agglomerate", True)

        # for k, v in six.iteritems(kwargs):
        #     self.cv_params[k] = v

        self.default_chunk = default_chunk

        self.obj = obj
        self.cvs = {}

        self.info = None
        self.fetch_info()

        if "info" in self.cv_params and overwrite:
            self.store_info()

    # def exists(self):
    #       s = Storage(self.path)
    #       return s.exists('info')

    def fetch_info(self):
        print("Fetching info")
        tmp_cv = self.obj(self.path, **self.cv_params)
        self.info = tmp_cv.info

    def serialize(self):
        contents = {
            "path": self.path,
            "allow_info_writes": self.allow_info_writes,
            "cv_params": self.cv_params,
        }
        s = json.dumps(contents)
        return s

    @classmethod
    def deserialize(cls, s, cache={}, **kwargs):
        if s in cache:
            return cache[s]
        else:
            import pdb

            pdb.set_trace()
            mcv = cls(s, **kwargs)
            cache[s] = mcv
            return mcv

    def get_info(self):
        return self.info

    def store_info(self, info=None):
        if not self.allow_info_writes:
            raise Exception(
                "Attempting to store info to {}, but "
                "'allow_info_writes' flag is set to False".format(self.path)
            )

        tmp_cv = self.obj(self.path, **self.cv_params)
        if info is not None:
            tmp_cv.info = info

        tmp_cv.commit_info()
        tmp_cv.commit_provenance()

    def ensure_info_has_mip(self, mip):
        tmp_cv = self.obj(self.path, **self.cv_params)
        scale_num = len(tmp_cv.info["scales"])

        if scale_num < mip + 1:
            while scale_num < mip + 1:
                tmp_cv.add_scale(
                    (2 ** scale_num, 2 ** scale_num, 1), chunk_size=self.default_chunk
                )
                scale_num += 1

            self.store_info(tmp_cv.info)

    def extend_info_to_mip(self, mip):
        info = self.get_info()
        highest_mip = len(info["scales"]) - 1
        highest_mip_info = info["scales"][-1]

        if highest_mip >= mip:
            return

        while highest_mip < mip:
            new_highest_mip_info = copy.deepcopy(highest_mip_info)
            # size
            # voxel offset
            # resolution -> key
            new_highest_mip_info["size"] = [
                highest_mip_info["size"][0] // 2,
                highest_mip_info["size"][1] // 2,
                highest_mip_info["size"][2],
            ]

            new_highest_mip_info["voxel_offset"] = [
                highest_mip_info["voxel_offset"][0] // 2,
                highest_mip_info["voxel_offset"][1] // 2,
                highest_mip_info["voxel_offset"][2],
            ]

            new_highest_mip_info["resolution"] = [
                highest_mip_info["resolution"][0] * 2,
                highest_mip_info["resolution"][1] * 2,
                highest_mip_info["resolution"][2],
            ]

            new_highest_mip_info["key"] = "_".join(
                [str(i) for i in new_highest_mip_info["resolution"]]
            )
            info["scales"].append(new_highest_mip_info)
            highest_mip += 1
            highest_mip_info = new_highest_mip_info

        self.store_info()

    def create(self, mip):
        corgie_logger.debug(
            "Creating CloudVolume for {0} at MIP{1}".format(self.path, mip)
        )
        self.extend_info_to_mip(mip)

        self.cvs[mip] = self.obj(self.path, mip=mip, **self.cv_params)

        # if self.mkdir:
        #  self.cvs[mip].commit_info()
        #  self.cvs[mip].commit_provenance()

    def __getitem__(self, mip):
        if mip not in self.cvs:
            self.create(mip)
        return self.cvs[mip]

    def __repr__(self):
        return self.path



import math
import torch
import copy

import click
import itertools
from click_option_group import optgroup

from corgie import scheduling
from corgie import helpers
from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords, BoundingCube
from corgie import argparsers
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
)


class ComputeStatsJob(scheduling.Job):
    LAYERS = {}

    def __init__(
        self,
        src_layer,
        mask_layers,
        mean_layer,
        var_layer,
        bcube,
        mip,
        chunk_xy,
        chunk_z,
    ):
        self.src_layer = src_layer
        self.mask_layers = mask_layers
        self.mean_layer = mean_layer
        self.var_layer = var_layer
        self.mip = mip
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z

        super().__init__()

    def task_generator(self):
        chunks = self.src_layer.break_bcube_into_chunks(
            bcube=self.bcube, chunk_xy=self.chunk_xy, chunk_z=self.chunk_z, mip=self.mip
        )

        assert len(chunks) % self.bcube.z_size() == 0
        chunks_per_section = len(chunks) // self.bcube.z_size()

        if (self.mean_layer.path, chunks_per_section) not in ComputeStatsJob.LAYERS:
            chunk_mean_layer = self.mean_layer.get_sublayer(
                name="chunk_mean",
                layer_type="section_value",
                num_channels=chunks_per_section)
            ComputeStatsJob.LAYERS[(self.mean_layer.path, chunks_per_section)] = chunk_mean_layer

        if (self.var_layer.path, chunks_per_section) not in ComputeStatsJob.LAYERS:
            chunk_var_layer = self.var_layer.get_sublayer(
                name="chunk_var",
                layer_type="section_value",
                num_channels=chunks_per_section)
            ComputeStatsJob.LAYERS[(self.var_layer.path, chunks_per_section)] = chunk_var_layer

        chunk_mean_layer = ComputeStatsJob.LAYERS[(self.mean_layer.path, chunks_per_section)]
        chunk_var_layer  = ComputeStatsJob.LAYERS[(self.var_layer.path, chunks_per_section)]

        chunks = self.src_layer.break_bcube_into_chunks(
            bcube=self.bcube,
            chunk_xy=self.chunk_xy,
            chunk_z=self.chunk_z,
            mip=self.mip,
            return_generator=True,
        )

        # legacy: chunks come sorted by z by default
        # chunks.sort(reverse=True, key=lambda c:c.z_range()[1])

        tasks = (
            ComputeStatsTask(
                self.src_layer,
                mean_layer=chunk_mean_layer,
                mask_layers=self.mask_layers,
                var_layer=chunk_var_layer,
                mip=self.mip,
                bcube=chunks[chunk_num],
                # chunks are sorted by z, so this gives the chunk num
                # for a given z
                write_channel=chunk_num % chunks_per_section,
            )
            for chunk_num in range(len(chunks))
        )

        corgie_logger.info(
            "Yielding chunk stats tasks: {}, MIP: {}".format(self.bcube, self.mip)
        )
        yield tasks
        yield scheduling.wait_until_done

        # the chunk mean values are written as different channels for 0, 0 pixel
        accum_bcube = copy.deepcopy(self.bcube)
        accum_bcube.reset_coords(xs=0, xe=1, ys=0, ye=1, mip=self.mip)

        accum_mean_chunks = chunk_mean_layer.break_bcube_into_chunks(
            bcube=accum_bcube,
            chunk_xy=chunks_per_section,
            chunk_z=self.chunk_z,
            mip=self.mip,
            return_generator=True,
        )

        accum_var_chunks = chunk_mean_layer.break_bcube_into_chunks(
            bcube=accum_bcube,
            chunk_xy=chunks_per_section,
            chunk_z=self.chunk_z,
            mip=self.mip,
            return_generator=True,
        )

        accum_mean_tasks = (
            ComputeStatsTask(
                chunk_mean_layer,
                mean_layer=self.mean_layer,
                var_layer=None,
                mip=self.mip,
                bcube=accum_chunk,
                write_channel=0,
            )
            for accum_chunk in accum_mean_chunks
        )
        corgie_logger.info("Yielding chunk stats aggregation tasks...")
        yield accum_mean_tasks

        accum_var_tasks = (
            ComputeStatsTask(
                chunk_var_layer,
                mean_layer=self.var_layer,
                var_layer=None,
                mip=self.mip,
                bcube=accum_chunk,
                write_channel=0,
            )
            for accum_chunk in accum_var_chunks
        )
        yield accum_var_tasks

    def create_dst_layers(self):
        return mean_layer, var_layer


class ComputeStatsTask(scheduling.Task):
    def __init__(
        self,
        src_layer,
        mean_layer,
        var_layer,
        mip,
        bcube,
        write_channel,
        mask_layers=[],
    ):
        super().__init__(self)
        self.src_layer = src_layer
        self.mask_layers = mask_layers
        self.mean_layer = mean_layer
        self.var_layer = var_layer
        self.mip = mip
        self.bcube = bcube
        self.write_channel = write_channel

    def execute(self):
        src_data = self.src_layer.read(bcube=self.bcube, mip=self.mip)

        mask_data = helpers.read_mask_list(
            mask_list=self.mask_layers, bcube=self.bcube, mip=self.mip
        )

        if mask_data is not None:
            src_data = src_data[mask_data == 0]
        src_data = src_data[src_data != 0]
        src_data = src_data[src_data == src_data]

        write_bcube = copy.deepcopy(self.bcube)
        write_bcube.reset_coords(xs=0, xe=1, ys=0, ye=1)

        if self.mean_layer is not None:
            if len(src_data) == 0:
                mean = torch.FloatTensor([0], device=src_data.device)
            else:
                mean = src_data.float().mean()

            self.mean_layer.write(
                mean,
                bcube=write_bcube,
                mip=self.mip,
                channel_start=self.write_channel,
                channel_end=self.write_channel + 1,
            )

        if self.var_layer is not None:
            if len(src_data) == 0:
                var = torch.FloatTensor([0], device=src_data.device)
            else:
                var = src_data.float().var()

            self.var_layer.write(
                var,
                bcube=write_bcube,
                mip=self.mip,
                channel_start=self.write_channel,
                channel_end=self.write_channel + 1,
            )


@click.command()
@corgie_optgroup("Layer parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where aligned stack will go",
)


# Other Params
@corgie_option("--suffix", "-s", nargs=1, type=str, default=None)
@corgie_optgroup("Computation Specification")
@corgie_option("--mip", "-m", nargs=1, type=int, required=True)
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=4096)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.pass_context
def compute_stats(
    ctx,
    src_layer_spec,
    dst_folder,
    suffix,
    mip,
    chunk_xy,
    chunk_z,
    start_coord,
    end_coord,
    coord_mip,
):
    compute_stats_fn(
        ctx,
        src_layer_spec,
        mask_layers_spec,
        suffix,
        mip,
        chunk_xy,
        chunk_z,
        start_coord,
        end_coord,
        coord_mip,
    )


def compute_stats_fn(
    ctx,
    src_layer_spec,
    dst_folder,
    suffix,
    mip,
    chunk_xy,
    chunk_z,
    start_coord,
    end_coord,
    coord_mip,
):

    if chunk_z != 1:
        raise NotImplemented(
            "Compute Statistics command currently only \
                supports per-section statistics."
        )

    scheduler = ctx.obj["scheduler"]

    src_layer = create_layer_from_spec(
        src_layer_spec, caller_name="src layer", readonly=True
    )

    mask_layers = [
        create_layer_from_spec(
            mask_spec,
            caller_name="src mask",
            readonly=True,
            allowed_types=["mask"],
            default_type="mask",
        )
        for mask_spec in mask_layers_spec
    ]

    if suffix is None:
        suffix = ""
    else:
        suffix = "_" + suffix

    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    mask_layers = src_stack.get_layers_of_type(["mask"])
    non_mask_layers = src_stack.get_layers_of_type(["img", "field"])

    for l in non_mask_layers:
        mean_layer = src_layer.get_sublayer(
            name=f"mean{suffix}",
            path=os.path.join(dst_dir, f"mean{suffix}"),
            layer_type="section_value",
        )

        var_layer = src_layer.get_sublayer(
            path=os.path.join(dst_dir, f"var{suffix}"), layer_type="section_value"
        )

        compute_stats_job = ComputeStatsJob(
            src_layer=l,
            mask_layers=mask_layers,
            mean_layer=mean_layer,
            var_layer=var_layer,
            bcube=bcube,
            mip=mip,
            chunk_xy=chunk_xy,
            chunk_z=chunk_z,
        )

        # create scheduler and execute the job
        scheduler.register_job(
            compute_stats_job, job_name=f"Compute Stats. Layer: {l}, Bcube: {bcube}"
        )
    scheduler.execute_until_completion()



# Stack class
# Stack of sections
# defined by:
# 3d bounding box
# cv
# section is a stack with thickness == 1
import copy
import os
import math

import six

from corgie import exceptions, helpers
from corgie.layers import str_to_layer_type
from corgie.log import logger as corgie_logger


class StackBase:
    def __init__(self, name=None):
        self.name = name
        self.layers = {}
        self.reference_layer = None

    def add_layer(self, layer):
        if layer.name is None:
            raise exceptions.UnnamedLayerException(
                layer,
                f"Layer name "
                f"needs to be set for it to be added to {self.name} stack.",
            )
        # if layer.name in self.layers:
        #    raise exceptions.ArgumentError(layer, f"Layer with name "
        #            f"'{layer.name}' added twice to '{self.name}' stack.")
        if self.reference_layer is None:
            self.reference_layer = layer
        self.layers[layer.name] = layer

    def remove_layer(self, layer_name):
        del self.layers[layer_name]

    def __contains__(self, m):
        return m in self.layers.keys()

    def __getitem__(self, m):
        return self.layers[m]

    def __len__(self):
        return len(self.layers)

    def read_data_dict(self, **index):
        result = {}
        for l in self.layers:
            result[l.name] = l.read(**index)
        raise NotImplementedError

    def write_data_dict(self, data_dict):
        raise NotImplementedError

    def get_layers(self):
        return list(self.layers.values())

    def get_layers_of_type(self, type_names):
        if isinstance(type_names, str):
            type_names = [type_names]

        types = tuple(str_to_layer_type(n) for n in type_names)

        result = []
        for k, v in six.iteritems(self.layers):
            if isinstance(v, types):
                result.append(v)

        return result

    def get_layer_types(self):
        layer_types = set()
        for k, v in six.iteritems(self.layers):
            layer_types.add(v.get_layer_type())

        return list(layer_types)


class Stack(StackBase):
    def __init__(self, name=None, layer_list=[], folder=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.folder = folder

        for l in layer_list:
            self.add_layer(l)

    def create_sublayer(
        self,
        name,
        layer_type,
        suffix="",
        reference=None,
        attach_to_stack=True,
        custom_folder=None,
        **kwargs,
    ):
        if (self.folder is None) and (custom_folder is None):
            raise exceptions.CorgieException(
                "Stack must have 'folder' field set " "before sublayers can be created"
            )

        if self.reference_layer is None and reference is None:
            raise exceptions.CorgieException(
                "Stack must either have at least one layer "
                "or reference layer must be provided for sublayer creation"
            )

        if reference is None:
            reference = self.reference_layer

        folder = custom_folder or self.folder
        path = os.path.join(folder, layer_type, f"{name}{suffix}")
        l = reference.backend.create_layer(
            path=path,
            layer_type=layer_type,
            name=name,
            reference=reference,
            **kwargs,
        )
        if attach_to_stack:
            self.add_layer(l)
        return l

    def create_unattached_sublayer(
        self, name, layer_type, suffix="", reference=None, **kwargs
    ):
        """
        Create a sublayer that does not get added to the stack.
        Make separate function for the sake of readability.
        """
        return self.create_sublayer(
            name,
            layer_type,
            suffix=suffix,
            reference=reference,
            attach_to_stack=False,
            **kwargs,
        )

    def read_data_dict(
        self,
        bcube,
        mip,
        translation_adjuster=None,
        stack_name=None,
        add_prefix=True,
        translation=None,
    ):

        data_dict = {}

        if translation is None:
            translation = helpers.Translation(0, 0)

        if stack_name is None:
            stack_name == self.name

        if stack_name is None or not add_prefix:
            name_prefix = ""
        else:
            name_prefix = f"{stack_name}_"

        agg_field = None
        field_layers = self.get_layers_of_type("field")
        # Assume that the last field is the final one
        if len(field_layers) > 0:
            l = field_layers[-1]
            global_name = "{}{}".format(name_prefix, l.name)
            this_field = l.read(bcube=bcube, mip=mip)
            data_dict[global_name] = this_field
            agg_field = this_field

        """for l in field_layers:
            global_name = "{}{}".format(name_prefix, l.name)
            this_field = l.read(bcube=bcube, mip=mip)
            data_dict[global_name] = this_field
            agg_field = this_field
            if agg_field is None:
                agg_field = this_field
            else:
                agg_field = this_field.field().from_pixels()(agg_field.field().from_pixels()).pixels()"""

        assert f"{name_prefix}agg_field" not in data_dict
        data_dict[f"{name_prefix}agg_field"] = agg_field

        if translation_adjuster is not None:
            src_field_trans = translation_adjuster(agg_field)
            translation.x += src_field_trans.x
            translation.y += src_field_trans.y

        # if translation.x != 0 or translation.y != 0:
        # import pdb; pdb.set_trace()
        final_bcube = copy.deepcopy(bcube)
        final_bcube = final_bcube.translate(
            x_offset=translation.y, y_offset=translation.x, mip=mip
        )

        for l in self.get_layers_of_type(["mask", "img"]):
            global_name = f"{name_prefix}{l.name}"
            data_dict[global_name] = l.read(bcube=final_bcube, mip=mip)
        return translation, data_dict

    def z_range(self):
        return self.bcube.z_range()

    def cutout(self):
        raise NotImplementedError


def create_stack_from_reference(
    reference_stack, folder, name, types=None, suffix="", **kwargs
):
    result = Stack(name=name, folder=folder)
    if types is None:
        layers = reference_stack.get_layers()
    else:
        layers = reference_stack.get_layers_of_type(types)

    for l in layers:
        result.create_sublayer(
            name=l.name,
            layer_type=l.get_layer_type(),
            suffix=suffix,
            reference=l,
            dtype=l.get_data_type(),
            **kwargs,
        )
    return result


class FieldSet:
    """Collection of Field layers to handle composition"""

    def __init__(self, layers=[]):
        self.layers = layers

    def append(self, layer):
        self.layers.append(layer)

    def get_field(self, layer, bcube, mip, dist, **kwargs):
        """Get field

        Args:
            layer (Layer)
            bcube (BoundingCube)
            mip (int)
            dist (float)

        Returns:
            TorchField
        """
        # corgie_logger.debug(f"get_field")
        # corgie_logger.debug(f"\tlayer={layer}")
        # corgie_logger.debug(f"\tbcube={bcube}")
        # corgie_logger.debug(f"\tmip={mip}")
        for k, v in kwargs.items():
            corgie_logger.debug(f"\t{k}={v}")

        f = layer.read(bcube=bcube, mip=mip, **kwargs).field_()
        return f  # * c

    def read(self, bcube, z_list, mip, **kwargs):
        """Compute composition of fields indexed by bcube & z_list

        This takes a list of fields, [f_0, f_1, ..., f_n],
        and composes them to get
        f_0 ⚬ f_1 ⚬ ... ⚬ f_n ~= f_0(f_1(...(f_n)))
        """
        if isinstance(z_list, int):
            z_list = [z_list] * len(self.layers)
        assert len(z_list) == len(self.layers)

        src_z = z_list[-1]
        z = z_list[0]
        layer = self.layers[0]
        abcube = bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
        agg_field = self.get_field(
            layer=layer, bcube=abcube, mip=mip, dist=src_z - z, **kwargs
        )
        for z, layer in zip(z_list[1:], self.layers[1:]):
            trans = helpers.percentile_trans_adjuster(agg_field)
            corgie_logger.debug(f"{trans} initial")
            trans = trans.round_to_mip(mip, layer.data_mip)
            corgie_logger.debug(f"{trans} round from M{mip} to M{layer.data_mip}")
            abcube = bcube.reset_coords(zs=z, ze=z + 1, in_place=False)
            abcube = abcube.translate(x_offset=trans.y, y_offset=trans.x, mip=mip)
            trans = trans.to_tensor(device=agg_field.device)
            agg_field -= trans
            agg_field = agg_field.from_pixels()
            this_field = self.get_field(
                layer=layer, bcube=abcube, mip=mip, dist=src_z - z, **kwargs
            )
            this_field = this_field.from_pixels()
            agg_field = agg_field(this_field)
            agg_field = agg_field.pixels()
            agg_field += trans
        return agg_field


class DistanceFieldSet(FieldSet):
    def __init__(self, decay_dist, layers=[]):
        """Compose set of fields, linearly scaling fields based on distance from source"""
        super().__init__(layers)
        self.decay_dist = decay_dist

    def get_field(self, layer, bcube, mip, dist, **kwargs):
        """Get field, adjusted by distance

        Args:
            layer (Layer)
            bcube (BoundingCube)
            mip (int)
            dist (float)

        Returns:
            TorchField adjusted (blurred/attenuated) by distance
        """
        # TODO: add ability to get blurred field using trilinear interpolation
        c = min(max(1.0 - (dist / self.decay_dist), 0.0), 1.0)
        corgie_logger.debug(f"get_field, c={c}")
        # TODO: ignore getting field if c==0
        # corgie_logger.debug(f"\tlayer={layer}")
        # corgie_logger.debug(f"\tbcube={bcube}")
        # corgie_logger.debug(f"\tmip={mip}")
        for k, v in kwargs.items():
            corgie_logger.debug(f"\t{k}={v}")

        f = layer.read(bcube=bcube, mip=mip, **kwargs).field_()
        return f * c


class PyramidDistanceFieldSet(FieldSet):
    def __init__(self, decay_dist, blur_rate, layers=[]):
        """Compose set of fields, linearly scaling & blurring fields based on distance
        from source.

        Requires that the field be downsampled into MIP-pyramid by factors of 2.

        NOTE: We are currently generating field pyramids with a box filter. We may want
        to change this to a Gaussian for a more precise filtering.

        Args:
            decay_dist (float): distance beyond which a field does not influence a neighbor
            blur_rate (float): rate of blurring increase with section distance. Can be roughly
                considered the change in size of std of gaussian kernel per single section (we
                currently create the MIP hierarchy with a box filter). For example, if the std 
                increases by 1 px every 10 sections, then blur_rate = 0.1.
        """
        super().__init__(layers)
        self.decay_dist = decay_dist
        self.blur_rate = blur_rate

    def get_field(self, layer, bcube, mip, dist, **kwargs):
        """Get field, adjusted by distance

        Args:
            layer (Layer)
            bcube (BoundingCube)
            mip (int)
            dist (float)

        Returns:
            TorchField adjusted (blurred & attenuated) by distance
        """
        c = min(max(1.0 - (dist / self.decay_dist), 0.0), 1.0)
        sigma_mip = math.log(dist * self.blur_rate + 1, 2)
        lower_mip = math.floor(sigma_mip)
        upper_mip = math.ceil(sigma_mip)
        corgie_logger.debug(
            f"get_field, c={c:.3f}, mip={sigma_mip+mip:.3f}, lower_mip={lower_mip+mip}, upper_mip={upper_mip+mip}"
        )
        if lower_mip == upper_mip:
            f = layer.read(bcube=bcube, mip=lower_mip + mip, **kwargs).field_()
        else:
            lf = layer.read(bcube=bcube, mip=lower_mip + mip, **kwargs).field_()
            uf = layer.read(bcube=bcube, mip=upper_mip + mip, **kwargs).field_()
            uf = uf.up(mips=1) * 2
            alpha = sigma_mip - lower_mip
            f = alpha * lf + (1 - alpha) * uf
        f = f.up(mips=lower_mip) * (2 ** lower_mip)
        return f * c



import click
import six

from corgie import scheduling, residuals, helpers, stack

from corgie.log import logger as corgie_logger
from corgie.layers import get_layer_types, DEFAULT_LAYER_TYPE, str_to_layer_type
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)


class RenderJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        dst_stack,
        mips,
        pad,
        render_masks,
        blackout_masks,
        bcube,
        chunk_xy,
        chunk_z,
        seethrough_mask_layer=None,
        additional_fields=[],
        seethrough_offset=-1,
        preserve_zeros=False,
    ):

        self.src_stack = src_stack
        self.dst_stack = dst_stack

        if isinstance(mips, int):
            mips = [mips]
        self.mips = mips

        self.pad = pad
        self.bcube = bcube
        self.chunk_xy = chunk_xy
        self.chunk_z = chunk_z
        self.render_masks = render_masks
        self.blackout_masks = blackout_masks
        self.additional_fields = additional_fields

        self.preserve_zeros = preserve_zeros
        self.seethrough_mask_layer = seethrough_mask_layer
        self.seethrough_offset = seethrough_offset

        super().__init__()

    def task_generator(self):
        for mip in self.mips:
            chunks = self.dst_stack.get_layers()[0].break_bcube_into_chunks(
                bcube=self.bcube,
                chunk_xy=self.chunk_xy,
                chunk_z=self.chunk_z,
                mip=mip,
                return_generator=True,
            )

            tasks = (
                RenderTask(
                    self.src_stack,
                    self.dst_stack,
                    blackout_masks=self.blackout_masks,
                    render_masks=self.render_masks,
                    mip=mip,
                    pad=self.pad,
                    bcube=input_chunk,
                    additional_fields=self.additional_fields,
                    preserve_zeros=self.preserve_zeros,
                    seethrough_mask_layer=self.seethrough_mask_layer,
                    seethrough_offset=self.seethrough_offset,
                )
                for input_chunk in chunks
            )
            corgie_logger.info(
                f"Yielding render tasks for bcube: {self.bcube}, MIP: {mip}"
            )

            yield tasks


class RenderTask(scheduling.Task):
    def __init__(
        self,
        src_stack,
        dst_stack,
        additional_fields,
        render_masks,
        blackout_masks,
        seethrough_offset,
        seethrough_mask_layer,
        mip,
        preserve_zeros,
        pad,
        bcube,
    ):
        super().__init__(self)
        self.src_stack = src_stack
        self.dst_stack = dst_stack
        self.render_masks = render_masks
        self.blackout_masks = blackout_masks
        self.mip = mip
        self.bcube = bcube
        self.pad = pad
        self.additional_fields = additional_fields
        self.preserve_zeros = preserve_zeros
        self.seethrough_offset = seethrough_offset
        self.seethrough_mask_layer = seethrough_mask_layer
        self.blackout_value = 0.0

    def execute(self):
        padded_bcube = self.bcube.uncrop(self.pad, self.mip)

        for f in self.additional_fields:
            # just in case the "additional" field is actually already a part of src_stack
            if f not in self.src_stack.layers.values():
                self.src_stack.add_layer(f)

        src_translation, src_data_dict = self.src_stack.read_data_dict(
            padded_bcube,
            mip=self.mip,
            add_prefix=False,
            translation_adjuster=helpers.percentile_trans_adjuster,
        )
        agg_field = src_data_dict[f"agg_field"]
        if agg_field is not None:
            agg_field[:, 0] -= src_translation.x
            agg_field[:, 1] -= src_translation.y

        # Produce and aggregate mask if there's a need to blackout all masks
        agg_mask = None
        if self.blackout_masks:
            mask_layers = self.dst_stack.get_layers_of_type(["mask"])
            mask_layer_names = [l.name for l in mask_layers]
            for n, d in six.iteritems(src_data_dict):
                if n in mask_layer_names:
                    if agg_mask is None:
                        agg_mask = d
                    else:
                        agg_mask = ((agg_mask + d) > 0).byte()

            # if the mask is not empty, warp it and coarsen it for seethrough
            if agg_mask is not None:
                coarsen_factor = int(2 ** (6 - self.mip))
                agg_mask = helpers.coarsen_mask(agg_mask, coarsen_factor)
                if agg_field is not None:
                    warped_mask = residuals.res_warp_img(agg_mask.float(), agg_field)
                else:
                    warped_mask = agg_mask

                warped_mask = (warped_mask > 0.4).byte()
                # To save time, we won't be blacking out stuff that gets cropped anyway
                warped_mask = helpers.crop(warped_mask, self.pad)
            else:
                warped_mask = None

        if self.render_masks:
            write_layers = self.dst_stack.get_layers_of_type(["img", "mask"])
        else:
            write_layers = self.dst_stack.get_layers_of_type("img")

        for l in write_layers:
            src = src_data_dict[f"{l.name}"]
            '''
            if (src != 0).sum():
                import pdb; pdb.set_trace()
            else:
                return
            '''
            if agg_field is not None:
                warped_src = residuals.res_warp_img(src.float(), agg_field)
            else:
                warped_src = src

            cropped_out = helpers.crop(warped_src, self.pad)

            if l.get_layer_type() == "img":
                if self.blackout_masks and warped_mask is not None:
                    cropped_out[warped_mask] = self.blackout_value

                if self.preserve_zeros and agg_field is not None:
                    src_zero_mask = src == 0
                    warped_zero_mask = residuals.res_warp_img(
                        src_zero_mask.float(), agg_field
                    )
                    warped_zero_mask = (warped_zero_mask > 0.4).byte()
                    warped_zero_mask = helpers.crop(warped_zero_mask, self.pad)
                    cropped_out[warped_zero_mask] = 0

                if self.seethrough_mask_layer is not None:
                    seethrough_mask = (
                        self.seethrough_mask_layer.read(bcube=self.bcube, mip=self.mip)
                        != 0
                    )
                    seethrough_bcube = self.bcube.translate(
                        z_offset=self.seethrough_offset
                    )
                    seethrough_data = l.read(bcube=seethrough_bcube, mip=self.mip)

                    cropped_out[seethrough_mask] = seethrough_data[seethrough_mask]

                    seenthru = (cropped_out[seethrough_mask] != 0).sum()
                    corgie_logger.debug(f"Seenthrough {seenthru} pixels")

            l.write(cropped_out, bcube=self.bcube, mip=self.mip)

        for f in self.additional_fields:
            # remove fields we added
            self.src_stack.remove_layer(f.name)


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
#
@corgie_option(
    "--dst_folder",
    nargs=1,
    type=str,
    required=True,
    help="Folder where rendered stack will go",
)
@corgie_optgroup("Render Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--chunk_z", nargs=1, type=int, default=1)
@corgie_option("--pad", nargs=1, type=int, default=512)
@corgie_option("--mip", "mips", nargs=1, type=int, required=True, multiple=True)
@corgie_option("--render_masks/--no_render_masks", default=True)
@corgie_option("--blackout_masks/--no_blackout_masks", default=False)
@corgie_option("--seethrough/--no_seethrough", default=False)
@corgie_option("--force_chunk_xy", nargs=1, type=int)
@corgie_option("--force_chunk_z", nargs=1, type=int)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@click.option("--suffix", nargs=1, type=str, default=None)
@click.pass_context
def render(
    ctx,
    src_layer_spec,
    dst_folder,
    pad,
    render_masks,
    blackout_masks,
    seethrough,
    chunk_xy,
    chunk_z,
    start_coord,
    end_coord,
    mips,
    coord_mip,
    force_chunk_xy,
    force_chunk_z,
    suffix,
):
    scheduler = ctx.obj["scheduler"]

    if suffix is None:
        suffix = "_rendered"
    else:
        suffix = f"_{suffix}"

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    if not force_chunk_xy:
        force_chunk_xy = chunk_xy

    if not force_chunk_z:
        force_chunk_z = chunk_z

    dst_stack = stack.create_stack_from_reference(
        reference_stack=src_stack,
        folder=dst_folder,
        name="dst",
        types=["img", "mask"],
        force_chunk_xy=force_chunk_xy,
        force_chunk_z=force_chunk_z,
        suffix=suffix,
        overwrite=True,
    )
    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    render_job = RenderJob(
        src_stack=src_stack,
        dst_stack=dst_stack,
        mips=mips,
        pad=pad,
        bcube=bcube,
        chunk_xy=chunk_xy,
        chunk_z=chunk_z,
        render_masks=render_masks,
        blackout_masks=blackout_masks,
    )

    # create scheduler and execute the job
    scheduler.register_job(render_job, job_name="Render {}".format(bcube))
    scheduler.execute_until_completion()



import click
import procspec
import math

from corgie import scheduling, helpers

from corgie.log import logger as corgie_logger
from corgie.boundingcube import get_bcube_from_coords
from corgie.argparsers import (
    LAYER_HELP_STR,
    create_layer_from_spec,
    corgie_optgroup,
    corgie_option,
    create_stack_from_spec,
)

from corgie.cli.common import ChunkedJob


class CompareSectionsJob(scheduling.Job):
    def __init__(
        self,
        src_stack,
        dst_layer,
        chunk_xy,
        processor_spec,
        mip,
        dst_mip,
        pad,
        bcube,
        tgt_z_offset,
        tgt_stack=None,
        suffix="",
    ):

        self.src_stack = src_stack
        if tgt_stack is None:
            tgt_stack = src_stack

        self.tgt_stack = tgt_stack
        self.dst_layer = dst_layer
        self.chunk_xy = chunk_xy
        self.pad = pad
        self.bcube = bcube
        self.tgt_z_offset = tgt_z_offset

        self.suffix = suffix

        self.processor_spec = processor_spec
        self.mip = mip
        self.dst_mip = dst_mip
        super().__init__()

    def task_generator(self):
        chunks = self.dst_layer.break_bcube_into_chunks(
            bcube=self.bcube,
            chunk_xy=self.chunk_xy,
            chunk_z=1,
            mip=self.dst_mip,
            return_generator=True,
        )

        tasks = (
            CompareSectionsTask(
                processor_spec=self.processor_spec,
                src_stack=self.src_stack,
                tgt_stack=self.tgt_stack,
                dst_layer=self.dst_layer,
                mip=self.mip,
                dst_mip=self.dst_mip,
                pad=self.pad,
                tgt_z_offset=self.tgt_z_offset,
                bcube=input_chunk,
            )
            for input_chunk in chunks
        )
        print(
            f"Yielding compare-sections tasks for bcube: {self.bcube}, MIP: {self.mip}"
        )

        yield tasks


class CompareSectionsTask(scheduling.Task):
    def __init__(
        self,
        processor_spec,
        src_stack,
        tgt_stack,
        dst_layer,
        mip,
        dst_mip,
        pad,
        tgt_z_offset,
        bcube,
    ):
        super().__init__()
        self.processor_spec = processor_spec
        self.src_stack = src_stack
        self.tgt_stack = tgt_stack
        self.dst_layer = dst_layer
        self.mip = mip
        self.dst_mip = dst_mip
        self.pad = pad
        self.tgt_z_offset = tgt_z_offset
        self.bcube = bcube

    def execute(self):
        src_bcube = self.bcube.uncrop(self.pad, self.mip)
        tgt_bcube = src_bcube.translate(z_offset=self.tgt_z_offset)
        processor = procspec.parse_proc(spec_str=self.processor_spec)

        _, tgt_data_dict = self.tgt_stack.read_data_dict(
            tgt_bcube, mip=self.mip, stack_name="tgt"
        )

        _, src_data_dict = self.src_stack.read_data_dict(
            src_bcube, mip=self.mip, stack_name="src"
        )

        processor_input = {**src_data_dict, **tgt_data_dict}

        result = processor(processor_input, output_key="result")
        crop = self.pad // 2 ** (self.dst_mip - self.mip)
        cropped_result = helpers.crop(result, crop)
        self.dst_layer.write(cropped_result, bcube=self.bcube, mip=self.dst_mip)


@click.command()
@corgie_optgroup("Layer Parameters")
@corgie_option(
    "--src_layer_spec",
    "-s",
    nargs=1,
    type=str,
    required=True,
    multiple=True,
    help="Source layer spec. Use multiple times to include all masks, fields, images. "
    + LAYER_HELP_STR,
)
@corgie_option(
    "--tgt_layer_spec",
    "-t",
    nargs=1,
    type=str,
    required=False,
    multiple=True,
    help="Target layer spec. Use multiple times to include all masks, fields, images. "
    "DEFAULT: Same as source layers",
)
@corgie_option(
    "--dst_layer_spec",
    nargs=1,
    type=str,
    required=True,
    help="Specification for the destination layer. Must be an image or mask type.",
)
@corgie_option("--reference_key", nargs=1, type=str, default="img")
@corgie_optgroup("Compare Sections Method Specification")
@corgie_option("--chunk_xy", "-c", nargs=1, type=int, default=1024)
@corgie_option("--force_chunk_xy", nargs=1, type=int)
@corgie_option(
    "--pad",
    nargs=1,
    type=int,
    default=512,
)
@corgie_option("--processor_spec", nargs=1, type=str, multiple=False, required=True)
@corgie_option("--mip", nargs=1, type=int, multiple=False, required=True)
@corgie_option(
    "--dst_mip",
    nargs=1,
    type=int,
    multiple=False,
    required=False,
    help="When the output of the model may be at a different resolution than the input",
)
@corgie_optgroup("Data Region Specification")
@corgie_option("--start_coord", nargs=1, type=str, required=True)
@corgie_option("--end_coord", nargs=1, type=str, required=True)
@corgie_option("--coord_mip", nargs=1, type=int, default=0)
@corgie_option("--tgt_z_offset", nargs=1, type=int, default=-1)
@click.option("--suffix", nargs=1, type=str, default=None)
@click.pass_context
def compare_sections(
    ctx,
    src_layer_spec,
    tgt_layer_spec,
    dst_layer_spec,
    suffix,
    processor_spec,
    pad,
    chunk_xy,
    force_chunk_xy,
    start_coord,
    mip,
    dst_mip,
    end_coord,
    coord_mip,
    tgt_z_offset,
    reference_key,
):
    if suffix is None:
        suffix = ""
    else:
        suffix = f"_{suffix}"

    if not force_chunk_xy:
        force_chunk_xy = chunk_xy

    if not dst_mip:
        dst_mip = mip

    scheduler = ctx.obj["scheduler"]

    corgie_logger.debug("Setting up layers...")
    src_stack = create_stack_from_spec(src_layer_spec, name="src", readonly=True)

    tgt_stack = create_stack_from_spec(
        tgt_layer_spec, name="tgt", readonly=True, reference=src_stack
    )

    reference_layer = None
    if reference_key in src_stack.layers:
        reference_layer = src_stack.layers[reference_key]

    dst_layer = create_layer_from_spec(
        dst_layer_spec,
        allowed_types=["img", "mask"],
        default_type="field",
        readonly=False,
        caller_name="dst_layer",
        reference=reference_layer,
        force_chunk_xy=force_chunk_xy,
        overwrite=True,
    )

    bcube = get_bcube_from_coords(start_coord, end_coord, coord_mip)

    compare_job = CompareSectionsJob(
        src_stack=src_stack,
        tgt_stack=tgt_stack,
        dst_layer=dst_layer,
        chunk_xy=chunk_xy,
        processor_spec=processor_spec,
        pad=pad,
        bcube=bcube,
        tgt_z_offset=tgt_z_offset,
        suffix=suffix,
        mip=mip,
        dst_mip=dst_mip,
    )

    # create scheduler and execute the job
    scheduler.register_job(
        compare_job,
        job_name="Compare Job {}, tgt z offset {}".format(bcube, tgt_z_offset),
    )
    scheduler.execute_until_completion()



# To add new commands, create a file in this folder implementing a command,
# import the command here and add it to the list:
from corgie.cli.downsample import downsample
from corgie.cli.upsample import upsample
from corgie.cli.compute_field import compute_field
from corgie.cli.compute_stats import compute_stats
from corgie.cli.normalize import normalize
from corgie.cli.align_block import align_block
from corgie.cli.align import align
from corgie.cli.render import render
from corgie.cli.copy import copy
from corgie.cli.apply_processor import apply_processor
from corgie.cli.invert_field import invert_field
from corgie.cli.create_skeletons import create_skeletons
from corgie.cli.transform_skeletons import transform_skeletons
from corgie.cli.filter_skeletons import filter_skeletons
from corgie.cli.merge_copy import merge_copy
from corgie.cli.compute_field_by_spec import compute_field_by_spec
from corgie.cli.merge_render import merge_render
from corgie.cli.apply_processor_by_spec import apply_processor_by_spec
from corgie.cli.normalize_by_spec import normalize_by_spec
from corgie.cli.downsample_by_spec import downsample_by_spec
from corgie.cli.compare_sections import compare_sections
from corgie.cli.seethrough import seethrough_block
from corgie.cli.fill_nearest import fill_nearest
from corgie.cli.vote import vote
from corgie.cli.combine_masks import combine_masks
from corgie.cli.create_segmentation_masks import create_segmentation_masks
from corgie.cli.multi_section_compare import multi_section_compare

COMMAND_LIST = []
COMMAND_LIST.append(downsample)
COMMAND_LIST.append(upsample)
COMMAND_LIST.append(compute_field)
COMMAND_LIST.append(compute_stats)
COMMAND_LIST.append(compare_sections)
COMMAND_LIST.append(normalize)
COMMAND_LIST.append(align_block)
COMMAND_LIST.append(align)
COMMAND_LIST.append(render)
COMMAND_LIST.append(copy)
COMMAND_LIST.append(apply_processor)
COMMAND_LIST.append(invert_field)
COMMAND_LIST.append(create_skeletons)
COMMAND_LIST.append(transform_skeletons)
COMMAND_LIST.append(filter_skeletons)
COMMAND_LIST.append(merge_copy)
COMMAND_LIST.append(compute_field_by_spec)
COMMAND_LIST.append(merge_render)
COMMAND_LIST.append(apply_processor_by_spec)
COMMAND_LIST.append(normalize_by_spec)
COMMAND_LIST.append(downsample_by_spec)
COMMAND_LIST.append(seethrough_block)
COMMAND_LIST.append(fill_nearest)
COMMAND_LIST.append(vote)
COMMAND_LIST.append(compare_sections)
COMMAND_LIST.append(combine_masks)
COMMAND_LIST.append(create_segmentation_masks)
COMMAND_LIST.append(multi_section_compare)


def get_command_list():
    return COMMAND_LIST
